import sys
from collections.abc import Iterable
from pathlib import Path
from typing import Any, Optional

from ape.utils._github import _GithubClient, github_client

if sys.version_info.minor >= 11:
    # 3.11 or greater
    # NOTE: type-ignore is for when running mypy on python versions < 3.11
    import tomllib  # type: ignore[import-not-found]
else:
    import toml as tomllib  # type: ignore[no-redef]

from yaml import safe_load

from ape.api.config import ApeConfig
from ape.api.projects import ProjectAPI
from ape.utils.os import expand_environment_variables


class BrownieProject(ProjectAPI):
    """
    Allows traditional Brownie projects to work with Ape.
    This class implements the necessary methods in order
    to detect config settings in a Brownie project and
    treat it like an Ape project.
    """

    @property
    def brownie_config_file(self) -> Path:
        return self.path / "brownie-config.yaml"

    @property
    def is_valid(self) -> bool:
        return self.brownie_config_file.is_file()

    def extract_config(self, **overrides) -> ApeConfig:
        migrated_config_data: dict[str, Any] = {}
        text = self.brownie_config_file.read_text()
        text = expand_environment_variables(text)

        try:
            brownie_config_data = safe_load(text) or {}
        except Exception:
            brownie_config_data = {}

        contracts_folder = brownie_config_data.get("contracts_folder", "contracts")
        migrated_config_data["contracts_folder"] = contracts_folder

        # Migrate dependencies
        dependencies = []
        for dependency in brownie_config_data.get("dependencies", []):
            dependency_dict = {}
            dep_parts = dependency.split("/")
            gh_name = dep_parts[0]
            dep_name = gh_name.lower()
            if len(dep_parts) > 1:
                dependency_dict["name"] = dep_name
                if "@" in dep_parts[1]:
                    suffix_parts = dep_parts[1].split("@")
                    dependency_dict["github"] = f"{gh_name}/{suffix_parts[0]}"
                    dependency_dict["version"] = suffix_parts[1]
                else:
                    dependency_dict["github"] = dep_parts[1]

            if dependency_dict:
                dependencies.append(dependency_dict)

        if dependencies:
            migrated_config_data["dependencies"] = dependencies

        # Migrate solidity remapping
        import_remapping = []
        solidity_version = None
        if "compiler" in brownie_config_data:
            compiler_config = brownie_config_data["compiler"]
            if "solc" in compiler_config:
                solidity_config = compiler_config["solc"]
                solidity_version = solidity_config.get("version")

                available_dependencies = [d["name"] for d in dependencies]
                brownie_import_remapping = solidity_config.get("remappings", [])

                for remapping in brownie_import_remapping:
                    parts = remapping.split("=")
                    map_key = parts[0]
                    real_path = parts[1]

                    real_path_parts = real_path.split("/")
                    dependency_name = real_path_parts[0].lower()

                    if dependency_name in available_dependencies:
                        suffix = real_path_parts[1]
                        if "@" in suffix:
                            version_id = suffix.split("@")[1]
                            entry = f"{dependency_name}/{version_id}"
                            import_remapping.append(f"{map_key}={entry}")
                        else:
                            import_remapping.append(f"{map_key}={dependency_name}")

        if import_remapping or solidity_version:
            migrated_solidity_config: dict[str, Any] = {}

            if import_remapping:
                migrated_solidity_config["import_remapping"] = import_remapping

            if solidity_version:
                migrated_solidity_config["version"] = solidity_version

            migrated_config_data["solidity"] = migrated_solidity_config

        model = {**migrated_config_data, **overrides}
        return ApeConfig.model_validate(model)


class FoundryProject(ProjectAPI):
    """
    Helps Ape read configurations from foundry projects
    and lessens the need of specifying ``config_override:``
    for foundry-based dependencies.
    """

    _github_client: _GithubClient = github_client

    @property
    def foundry_config_file(self) -> Path:
        return self.path / "foundry.toml"

    @property
    def gitmodules_file(self) -> Path:
        return self.path / ".gitmodules"

    @property
    def remapping_file(self) -> Path:
        return self.path / "remapping.txt"

    @property
    def is_valid(self) -> bool:
        return self.foundry_config_file.is_file()

    def extract_config(self, **overrides) -> "ApeConfig":
        ape_cfg: dict = {}

        root_data = self._parse_foundry_toml()

        # Handle root project configuration.
        # NOTE: The default contracts folder name is `src` in foundry
        #  instead of `contracts`, hence the default.
        contracts_folder = root_data.get("src")

        if not contracts_folder:
            if (self.foundry_config_file.parent / "src").is_dir():
                contracts_folder = "src"
            elif (self.foundry_config_file.parent / "contracts").is_dir():
                contracts_folder = "contracts"

        if contracts_folder:
            ape_cfg["contracts_folder"] = contracts_folder

        # Foundry uses git-submodules for dependencies.
        if dependencies := self._parse_dependencies_from_gitmodules():
            ape_cfg["dependencies"] = dependencies

        lib_paths = root_data.get("libs", ("lib",))
        if solidity := self._parse_solidity_config(
            root_data, dependencies, lib_paths, contracts_folder=contracts_folder
        ):
            ape_cfg["solidity"] = solidity

        return ApeConfig.model_validate({**ape_cfg, **overrides})

    def _parse_foundry_toml(self) -> dict:
        data = tomllib.loads(self.foundry_config_file.read_text())
        profile = data.get("profile", {})
        return profile.get("default", {})

    def _parse_dependencies_from_gitmodules(self) -> list[dict]:
        if not self.gitmodules_file.is_file():
            return []

        dependencies: list[dict] = []
        module_data = self._parse_gitmodules()
        for module in module_data:
            if not (url := module.get("url")):
                continue
            elif not url.startswith("https://github.com/"):
                # Not from GitHub.
                continue

            github = url.replace("https://github.com/", "").replace(".git", "")
            dependency = {"github": github}
            version_type, version = self._parse_version_from_module(module)
            dependency[version_type] = version
            dependency["name"] = github.split("/")[-1].lower().replace("_", "-")
            dependencies.append(dependency)

        return dependencies

    def _parse_gitmodules(self) -> list[dict[str, str]]:
        submodules: list[dict[str, str]] = []
        submodule: dict[str, str] = {}
        content = Path(self.gitmodules_file).read_text()

        for line in content.splitlines():
            line = line.strip()
            if line.startswith("[submodule"):
                # Add the submodule we have been building to the list
                # if it exists. This happens on submodule after the first one.
                if submodule:
                    submodules.append(submodule)
                    submodule = {}

            for key in ("path", "url", "release", "branch"):
                if not line.startswith(f"{key} ="):
                    continue

                submodule[key] = line.split("=")[1].strip()
                break  # No need to try the rest.

        # Add the last submodule.
        if submodule:
            submodules.append(submodule)

        return submodules

    def _parse_version_from_module(
        self, module: dict, default_version: str = "main"
    ) -> tuple[str, str]:
        if "release" in module:
            # Use GitHub version API.
            return ("version", module["release"])

        elif "branch" in module:
            # Use clone-by-reference.
            return ("ref", module["branch"])

        elif "url" not in module:
            return ("ref", default_version)

        url = module["url"]
        github = url.replace("https://github.com/", "").replace(".git", "")
        gh_parts = github.split("/")
        if len(gh_parts) != 2:
            # Likely not possible, but just try `main`.
            return ("ref", default_version)

        # Use the default branch of the repo.
        org_name, repo_name = github.split("/")
        repo = self._github_client.get_repo(org_name, repo_name)
        return ("ref", repo.get("default_branch", default_version))

    def _parse_solidity_config(
        self,
        data: dict,
        dependencies: list[dict],
        lib_paths: Iterable[str],
        contracts_folder: Optional[str] = None,
    ) -> dict:
        sol_cfg: dict = {}

        # Different foundry versions use a different key for the solc version.
        if version := (data.get("solc") or data.get("solc_version")):
            sol_cfg["version"] = version

        if evm_version := data.get("evm_version"):
            sol_cfg["evm_version"] = evm_version

        if via_ir := data.get("via_ir"):
            sol_cfg["via_ir"] = via_ir

        foundry_remappings = [*data.get("remappings", []), *self._parse_remappings_file()]
        if remappings := self._parse_remappings(
            foundry_remappings, dependencies, lib_paths, contracts_folder=contracts_folder
        ):
            sol_cfg["import_remapping"] = remappings

        return sol_cfg

    def _parse_remappings_file(self) -> list[str]:
        if not self.remapping_file.is_file():
            return []

        return self.remapping_file.read_text(encoding="utf8").splitlines()

    def _parse_remappings(
        self,
        foundry_remappings: list[str],
        dependencies: list[dict],
        lib_paths: Iterable[str],
        contracts_folder: Optional[str] = None,
    ) -> list[str]:
        ape_sol_remappings: set[str] = set()

        for f_remap in foundry_remappings:
            key, value = f_remap.split("=")
            sep = "\\" if "\\" in value else "/"
            real_key = key.rstrip(sep)
            clean_key = real_key.lstrip("@")

            # Check if is from one of the dependencies.
            is_dep = False
            repo = value
            for lib_path in lib_paths:
                if not value.startswith(f"{lib_path}{sep}"):
                    continue

                # Dependency found.
                is_dep = True
                repo = value.replace(f"{lib_path}{sep}", "").strip(sep)
                break

            if not is_dep:
                # Append as-is.
                ape_sol_remappings.add(f_remap)
                continue

            # Setup remapping to a dependency in a way Ape expects.
            # Also, change the name of the dependencies to be the short name
            # from the remapping (also what Ape expects).
            dep_found = False
            for dependency in dependencies:
                # NOTE: There appears to be no rhyme or reason to
                # dependency short-names in foundry.
                if (
                    not dependency["github"].endswith(clean_key)
                    and dependency["name"] != clean_key
                    and dependency["name"] != repo
                ):
                    continue

                # Matching dependency found.
                dependency["name"] = clean_key
                version = dependency.get("version") or dependency.get("ref")
                prefix = f"{sep}.cache{sep}"
                if contracts_folder:
                    prefix = f"{contracts_folder}{prefix}"

                value_without_lib_path = value
                for lib_path in lib_paths:
                    if f"{lib_path}{sep}" not in value:
                        continue

                    value_without_lib_path = value.replace(f"{lib_path}{sep}", "")

                # Sometimes, contracts-folder name is included.
                suffix = ""
                if f"{clean_key}{sep}" in value_without_lib_path:
                    suffix = value_without_lib_path.replace(f"{clean_key}{sep}", "").rstrip(sep)

                new_value = f"{prefix}{clean_key}{sep}{version}{sep}{suffix}"
                new_remapping = f"{real_key}={new_value}"
                ape_sol_remappings.add(new_remapping)
                dep_found = True
                break

            if not dep_found:
                # Item seems like a dependency but not found in `dependencies`.
                ape_sol_remappings.add(f_remap)

        return sorted(list(ape_sol_remappings))



from typing import TYPE_CHECKING, NewType, Optional, Union

from pydantic import NonNegativeInt, field_validator
from pydantic_settings import SettingsConfigDict

from ape.api.config import PluginConfig
from ape.utils.basemodel import ManagerAccessMixin
from ape.utils.testing import (
    DEFAULT_NUMBER_OF_TEST_ACCOUNTS,
    DEFAULT_TEST_ACCOUNT_BALANCE,
    DEFAULT_TEST_CHAIN_ID,
    DEFAULT_TEST_HD_PATH,
    DEFAULT_TEST_MNEMONIC,
)

if TYPE_CHECKING:
    from ape.pytest.utils import Scope


class EthTesterProviderConfig(PluginConfig):
    chain_id: int = DEFAULT_TEST_CHAIN_ID
    auto_mine: bool = True
    model_config = SettingsConfigDict(extra="allow", env_prefix="APE_TEST_")


class GasExclusion(PluginConfig):
    contract_name: str = "*"  # If only given method, searches across all contracts.
    method_name: Optional[str] = None  # By default, match all methods in a contract
    model_config = SettingsConfigDict(extra="allow", env_prefix="APE_TEST_")


CoverageExclusion = NewType("CoverageExclusion", GasExclusion)


class GasConfig(PluginConfig):
    """
    Configuration related to test gas reports.
    """

    exclude: list[GasExclusion] = []
    """
    Contract methods patterns to skip. Specify ``contract_name:`` and not
    ``method_name:`` to skip all methods in the contract. Only specify
    ``method_name:`` to skip all methods across all contracts. Specify
    both to skip methods in a certain contracts. Entries use glob-rules;
    use ``prefix_*`` to skip all items with a certain prefix.
    """

    reports: list[str] = []
    """
    Report-types to use. Currently, only supports `terminal`.
    """

    model_config = SettingsConfigDict(extra="allow", env_prefix="APE_TEST_")

    @field_validator("reports", mode="before")
    @classmethod
    def validate_reports(cls, values):
        values = list(set(values or []))
        valid = ("terminal",)
        for val in values:
            if val not in valid:
                valid_str = ", ".join(valid)
                raise ValueError(f"Invalid gas-report format '{val}'. Valid: {valid_str}")

        return values

    @property
    def show(self) -> bool:
        return "terminal" in self.reports


_ReportType = Union[bool, dict]
"""Dict is for extra report settings."""


class CoverageReportsConfig(PluginConfig):
    """
    Enable reports.
    """

    terminal: _ReportType = True
    """
    Set to ``False`` to hide the terminal coverage report.
    """

    xml: _ReportType = False
    """
    Set to ``True`` to generate an XML coverage report in your .build folder.
    """

    html: _ReportType = False
    """
    Set to ``True`` to generate HTML coverage reports.
    """

    model_config = SettingsConfigDict(extra="allow", env_prefix="APE_TEST_")

    @property
    def has_any(self) -> bool:
        return any(x not in ({}, None, False) for x in (self.html, self.terminal, self.xml))


class CoverageConfig(PluginConfig):
    """
    Configuration related to contract coverage.
    """

    track: bool = False
    """
    Setting this to ``True`` is the same as always running with
    the ``--coverage`` flag.
    """

    reports: CoverageReportsConfig = CoverageReportsConfig()
    """
    Enable reports.
    """

    exclude: list[CoverageExclusion] = []
    """
    Contract methods patterns to skip. Specify ``contract_name:`` and not
    ``method_name:`` to skip all methods in the contract. Only specify
    ``method_name:`` to skip all methods across all contracts. Specify
    both to skip methods in a certain contracts. Entries use glob-rules;
    use ``prefix_*`` to skip all items with a certain prefix.
    """

    model_config = SettingsConfigDict(extra="allow", env_prefix="APE_TEST_")


class IsolationConfig(PluginConfig):
    enable_session: bool = True
    """
    Set to ``False`` to disable session isolation.
    """

    enable_package: bool = True
    """
    Set to ``False`` to disable package isolation.
    """

    enable_module: bool = True
    """
    Set to ``False`` to disable module isolation.
    """

    enable_class: bool = True
    """
    Set to ``False`` to disable class isolation.
    """

    enable_function: bool = True
    """
    Set to ``False`` to disable function isolation.
    """

    model_config = SettingsConfigDict(extra="allow", env_prefix="APE_TEST_")

    def get_isolation(self, scope: "Scope") -> bool:
        return getattr(self, f"enable_{scope.name.lower()}")


class ApeTestConfig(PluginConfig):
    balance: int = DEFAULT_TEST_ACCOUNT_BALANCE
    """
    The starting-balance of every test account in Wei (NOT Ether).
    """

    coverage: CoverageConfig = CoverageConfig()
    """
    Configuration related to coverage reporting.
    """

    enable_fixture_rebasing: bool = True
    """
    Set to ``False`` to ignore fixture rebasing when non-function
    scoped fixtures become invalidated.
    """

    disconnect_providers_after: bool = True
    """
    Set to ``False`` to keep providers connected at the end of the test run.
    """

    gas: GasConfig = GasConfig()
    """
    Configuration related to gas reporting.
    """

    hd_path: str = DEFAULT_TEST_HD_PATH
    """
    The hd_path to use when generating the test accounts.
    """

    mnemonic: str = DEFAULT_TEST_MNEMONIC
    """
    The mnemonic to use when generating the test accounts.
    """

    number_of_accounts: NonNegativeInt = DEFAULT_NUMBER_OF_TEST_ACCOUNTS
    """
    The number of test accounts to generate in the provider.
    """

    provider: EthTesterProviderConfig = EthTesterProviderConfig()
    """
    Settings for the provider.
    """

    show_internal: bool = False
    """
    Set to ``True`` to always show Ape's internal stack-trace in errors,
    useful for debugging the framework itself.
    """

    isolation: Union[bool, IsolationConfig] = True
    """
    Configure which scope-specific isolation to enable. Set to
    ``False`` to disable all and ``True`` (default) to disable all.
    """

    model_config = SettingsConfigDict(extra="allow", env_prefix="APE_TEST_")

    @field_validator("balance", mode="before")
    @classmethod
    def validate_balance(cls, value):
        return (
            value
            if isinstance(value, int)
            else ManagerAccessMixin.conversion_manager.convert(value, int)
        )

    def get_isolation(self, scope: "Scope") -> bool:
        return (
            self.isolation
            if isinstance(self.isolation, bool)
            else self.isolation.get_isolation(scope)
        )



import json
from pathlib import Path

import pytest

from ape import Contract
from ape.contracts import ContractInstance
from ape.exceptions import ChainError

"""NOTE: This is testing Contract with a capital C."""


def test_Contract_from_abi(contract_instance):
    contract = Contract(contract_instance.address, abi=contract_instance.contract_type.abi)
    assert isinstance(contract, ContractInstance)
    assert contract.address == contract_instance.address
    assert contract.myNumber() == 0
    assert contract.balance == 0


def test_Contract_from_abi_list(contract_instance):
    contract = Contract(
        contract_instance.address,
        abi=[abi.model_dump() for abi in contract_instance.contract_type.abi],
    )

    assert isinstance(contract, ContractInstance)
    assert contract.address == contract_instance.address
    assert contract.myNumber() == 0


def test_Contract_from_json_str(contract_instance):
    contract = Contract(
        contract_instance.address,
        abi=json.dumps([abi.model_dump() for abi in contract_instance.contract_type.abi]),
    )

    assert isinstance(contract, ContractInstance)
    assert contract.address == contract_instance.address
    assert contract.myNumber() == 0


def test_Contract_from_json_str_retrieval_check_fails(mocker, chain, vyper_contract_instance):
    """
    Tests a bug when providing an abi= but fetch-attempt raises that we don't
    raise since the abi was already given.
    """
    # Make `.get()` fail.
    orig = chain.contracts.get
    mock_get = mocker.MagicMock()
    mock_get.side_effect = Exception

    abi_str = json.dumps([abi.model_dump() for abi in vyper_contract_instance.contract_type.abi])

    chain.contracts.get = mock_get
    try:
        contract = Contract(vyper_contract_instance.address, abi=abi_str)
    finally:
        chain.contracts.get = orig

    # Mostly, we are asserting it did not fail.
    assert isinstance(contract, ContractInstance)


def test_Contract_from_file(contract_instance):
    """
    need feedback about the json file specifications
    """
    PROJECT_PATH = Path(__file__).parent
    CONTRACTS_FOLDER = PROJECT_PATH / "data" / "contracts" / "ethereum" / "abi"
    json_abi_file = f"{CONTRACTS_FOLDER}/contract_abi.json"

    address = contract_instance.address
    contract = Contract(address, abi=json_abi_file)

    assert isinstance(contract, ContractInstance)
    assert contract.address == address
    assert contract.myNumber() == 0


def test_Contract_at_unknown_address(networks_connected_to_tester, address):
    _ = networks_connected_to_tester  # Need fixture or else get ProviderNotConnectedError
    with pytest.raises(ChainError, match=f"Failed to get contract type for address '{address}'."):
        Contract(address)


def test_Contract_specify_contract_type(
    vyper_contract_instance, solidity_contract_type, owner, networks_connected_to_tester
):
    # Solidity's contract type is very close to Vyper's.
    # This test purposely uses the other just to show we are able to specify it externally.
    contract = Contract(vyper_contract_instance.address, contract_type=solidity_contract_type)
    assert contract.address == vyper_contract_instance.address

    abis = [abi.name for abi in contract.contract_type.abi if hasattr(abi, "name")]
    assert "setNumber" in abis  # Shared ABI.
    assert "ACustomError" in abis  # SolidityContract-defined ABI.



from typing import Optional

from pydantic_settings import SettingsConfigDict

from ape.api.config import PluginConfig


class CustomNetwork(PluginConfig):
    """
    A custom network config.
    """

    name: str
    """Name of the network e.g. mainnet."""

    chain_id: int
    """Chain ID (required)."""

    ecosystem: str
    """The name of the ecosystem."""

    base_ecosystem_plugin: Optional[str] = None
    """The base ecosystem plugin to use, when applicable. Defaults to the default ecosystem."""

    default_provider: str = "node"
    """The default provider plugin to use. Default is the default node provider."""

    request_header: dict = {}
    """The HTTP request header."""

    model_config = SettingsConfigDict(extra="allow", env_prefix="APE_NETWORKS_")

    @property
    def is_fork(self) -> bool:
        """
        ``True`` when the name of the network ends in ``"-fork"``.
        """
        return self.name.endswith("-fork")


class NetworksConfig(PluginConfig):
    custom: list[CustomNetwork] = []
    model_config = SettingsConfigDict(extra="allow", env_prefix="APE_NETWORKS_")



import sys
from pathlib import Path

from ape.exceptions import ConfigError


def pytest_addoption(parser):
    def add_option(*names, **kwargs):
        try:
            parser.addoption(*names, **kwargs)
        except ValueError as err:
            name_str = ", ".join(names)
            if "already added" in str(err):
                raise ConfigError(
                    "Another pytest plugin besides `ape_test` uses an option with "
                    f"one of '{name_str}'. Note that Ape does not support being "
                    "installed alongside Brownie; please use separate environments!"
                )

            raise ConfigError(f"Failed adding option {name_str}: {err}") from err

    add_option("--show-internal", action="store_true")
    add_option(
        "--network",
        action="store",
        help="Override the default network and provider (see ``ape networks list`` for options).",
    )
    add_option(
        "--interactive",
        "-I",
        action="store_true",
        help="Open an interactive console each time a test fails.",
    )
    add_option(
        "--disable-isolation",
        action="store_true",
        help="Disable test and fixture isolation (see provider for info on snapshot availability).",
    )
    add_option(
        "--gas",
        action="store_true",
        help="Show a transaction gas report at the end of the test session.",
    )
    add_option(
        "--gas-exclude",
        action="store",
        help="A comma-separated list of contract:method-name glob-patterns to ignore.",
    )
    add_option("--coverage", action="store_true", help="Collect contract coverage.")
    add_option("--project", action="store", help="Change Ape's project")

    # NOTE: Other pytest plugins, such as hypothesis, should integrate with pytest separately


def pytest_configure(config):
    # Do not include ape internals in tracebacks unless explicitly asked
    if not config.getoption("--show-internal"):
        if path_str := sys.modules["ape"].__file__:
            base_path = str(Path(path_str).parent)

            def is_module(v):
                return getattr(v, "__file__", None) and v.__file__.startswith(base_path)

            for module in (v for v in sys.modules.values() if is_module(v)):
                # NOTE: Using try/except w/ type:ignore (over checking for attr)
                #   for performance reasons!
                try:
                    module.__tracebackhide__ = True  # type: ignore[attr-defined]
                except AttributeError:
                    pass

    if "--help" in config.invocation_params.args:
        # perf: Don't bother setting up runner if only showing help.
        return

    from ape.pytest.config import ConfigWrapper
    from ape.pytest.coverage import CoverageTracker
    from ape.pytest.fixtures import (
        FixtureManager,
        IsolationManager,
        PytestApeFixtures,
        ReceiptCapture,
    )
    from ape.pytest.gas import GasTracker
    from ape.pytest.runners import PytestApeRunner
    from ape.utils.basemodel import ManagerAccessMixin

    if project := config.getoption("--project"):
        ManagerAccessMixin.local_project.chdir(project)

    # Register the custom Ape test runner
    config_wrapper = ConfigWrapper(config)
    receipt_capture = ReceiptCapture(config_wrapper)
    isolation_manager = IsolationManager(config_wrapper, receipt_capture)
    fixture_manager = FixtureManager(config_wrapper, isolation_manager)
    gas_tracker = GasTracker(config_wrapper)
    coverage_tracker = CoverageTracker(config_wrapper)
    runner = PytestApeRunner(
        config_wrapper,
        isolation_manager,
        receipt_capture,
        gas_tracker,
        coverage_tracker,
        fixture_manager=fixture_manager,
    )
    config.pluginmanager.register(runner, "ape-test")

    # Inject runner for access to gas and coverage trackers.
    ManagerAccessMixin._test_runner = runner

    # Include custom fixtures for project, accounts etc.
    fixtures = PytestApeFixtures(config_wrapper, isolation_manager)
    config.pluginmanager.register(fixtures, "ape-fixtures")

    # Add custom markers
    config.addinivalue_line(
        "markers", "use_network(choice): Run this test using the given network choice."
    )



from collections import defaultdict
from collections.abc import Iterable, Iterator, Sequence
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union

from eth_pydantic_types import HexBytes

from ape.contracts import ContractContainer
from ape.exceptions import CompilerError, ContractLogicError, CustomError
from ape.logging import logger
from ape.managers.base import BaseManager
from ape.utils.basemodel import (
    ExtraAttributesMixin,
    ExtraModelAttributes,
    get_attribute_with_extras,
    only_raise_attribute_error,
)
from ape.utils.misc import log_instead_of_fail
from ape.utils.os import get_full_extension

if TYPE_CHECKING:
    from ethpm_types.contract_type import ContractType
    from ethpm_types.source import Content

    from ape.api.compiler import CompilerAPI
    from ape.managers.project import ProjectManager


class CompilerManager(BaseManager, ExtraAttributesMixin):
    """
    The singleton that manages :class:`~ape.api.compiler.CompilerAPI` instances.
    Each compiler plugin typically contains a single :class:`~ape.api.compiler.CompilerAPI`.

    **NOTE**: Typically, users compile their projects using the CLI via ``ape compile``,
    which uses the :class:`~ape.api.compiler.CompilerAPI` under-the-hood.

    Usage example::

        from ape import compilers  # "compilers" is the CompilerManager singleton
    """

    _registered_compilers_cache: dict[Path, dict[str, "CompilerAPI"]] = {}

    @log_instead_of_fail(default="<CompilerManager>")
    def __repr__(self) -> str:
        num_compilers = len(self.registered_compilers)
        cls_name = getattr(type(self), "__name__", CompilerManager.__name__)
        return f"<{cls_name} len(registered_compilers)={num_compilers}>"

    def __ape_extra_attributes__(self) -> Iterator[ExtraModelAttributes]:
        yield ExtraModelAttributes(
            name="compilers",
            # Allow referencing compilers by name e.g. `compilers.vyper`.
            attributes=lambda: {c.name: c for c in self.registered_compilers.values()},
        )

    @only_raise_attribute_error
    def __getattr__(self, attr_name: str) -> Any:
        return get_attribute_with_extras(self, attr_name)

    @cached_property
    def registered_compilers(self) -> dict[str, "CompilerAPI"]:
        """
        Each compile-able file extension mapped to its respective
        :class:`~ape.api.compiler.CompilerAPI` instance.

        Returns:
            dict[str, :class:`~ape.api.compiler.CompilerAPI`]: The mapping of file-extensions
            to compiler API classes.
        """
        registered_compilers = {}

        for plugin_name, (extensions, compiler_class) in self.plugin_manager.register_compiler:
            self.config_manager.get_config(plugin_name)
            compiler = compiler_class()

            for extension in extensions:
                if extension not in registered_compilers:
                    registered_compilers[extension] = compiler

        return registered_compilers

    def get_compiler(self, name: str, settings: Optional[dict] = None) -> Optional["CompilerAPI"]:
        for compiler in self.registered_compilers.values():
            if compiler.name != name:
                continue

            if settings is not None and settings != compiler.compiler_settings:
                # Use a new instance to support multiple compilers of same type.
                return compiler.model_copy(update={"compiler_settings": settings})

            return compiler

        return None

    def compile(
        self,
        contract_filepaths: Union[Path, str, Iterable[Union[Path, str]]],
        project: Optional["ProjectManager"] = None,
        settings: Optional[dict] = None,
        excluded_compilers: Optional[list[str]] = None,
    ) -> Iterator["ContractType"]:
        """
        Invoke :meth:`ape.ape.compiler.CompilerAPI.compile` for each of the given files.
        For example, use the `ape-solidity plugin <https://github.com/ApeWorX/ape-solidity>`__
        to compile ``'.sol'`` files.

        Raises:
            :class:`~ape.exceptions.CompilerError`: When there is no compiler found for the given
              file-extension as well as when there are contract-type collisions across compilers.

        Args:
            contract_filepaths (Union[Path, str, Iterable[Union[Path, str]]]): The files to
              compile, as ``pathlib.Path`` objects or path-strs.
            project (Optional[:class:`~ape.managers.project.ProjectManager`]): Optionally
              compile a different project that the one from the current-working directory.
            settings (Optional[Dict]): Adhoc compiler settings. Defaults to None.
              Ensure the compiler name key is present in the dict for it to work.

        Returns:
            Iterator[``ContractType``]: An iterator of contract types.
        """

        pm = project or self.local_project
        files_by_ext = defaultdict(list)
        if isinstance(contract_filepaths, (str, Path)):
            contract_filepaths = (contract_filepaths,)

        for path in map(Path, contract_filepaths):
            suffix = get_full_extension(path)
            if suffix in self.registered_compilers:
                files_by_ext[suffix].append(path)

        errors = []
        tracker: dict[str, str] = {}
        settings = settings or {}

        for next_ext, path_set in files_by_ext.items():
            compiler = self.registered_compilers[next_ext]
            if excluded_compilers and compiler.name.lower() in excluded_compilers:
                continue
            try:
                compiler_settings = settings.get(compiler.name, {})
                for contract in compiler.compile(path_set, project=pm, settings=compiler_settings):
                    if contract.name in tracker:
                        raise CompilerError(
                            f"ContractType collision. "
                            f"Contracts '{tracker[contract.name]}' and '{contract.source_id}' "
                            f"share the name '{contract.name}'."
                        )

                    if contract.name and contract.source_id:
                        tracker[contract.name] = contract.source_id

                    yield contract

            except CompilerError as err:
                # One of the compilers failed. Show the error but carry on.
                logger.log_debug_stack_trace()
                errors.append(err)
                continue

        if len(errors) == 1:
            # If only 1 error, just raise that.
            raise errors[0]

        elif len(errors) > 1:
            # Raise a combined error.
            formatted_errors = [f"{e}" for e in errors]
            error_message = "\n\n".join(formatted_errors)
            raise CompilerError(error_message)

        # else: successfully compiled everything!

    def compile_source(
        self,
        compiler_name: str,
        code: str,
        project: Optional["ProjectManager"] = None,
        settings: Optional[dict] = None,
        **kwargs,
    ) -> ContractContainer:
        """
        Compile the given program.

        Usage example::

            code = '[{"name":"foo","type":"fallback", "stateMutability":"nonpayable"}]'
            contract_container = compilers.compile_source(
                "ethpm",
                code,
                contractName="MyContract",
            )

        Args:
            compiler_name (str): The name of the compiler to use.
            code (str): The source code to compile.
            project (Optional[:class:`~ape.managers.project.ProjectManager`]): Optionally
              compile a different project that the one from the current-working directory.
            settings (Optional[dict]): Compiler settings.
            **kwargs (Any): Additional overrides for the ``ethpm_types.ContractType`` model.

        Returns:
            ``ContractContainer``: A contract container ready to be deployed.
        """
        compiler = self.get_compiler(compiler_name, settings=settings)
        if not compiler:
            raise ValueError(f"Compiler '{compiler_name}' not found.")

        contract_type = compiler.compile_code(code, project=project, **kwargs)
        return ContractContainer(contract_type=contract_type)

    def get_imports(
        self,
        contract_filepaths: Sequence[Path],
        project: Optional["ProjectManager"] = None,
    ) -> dict[str, list[str]]:
        """
        Combine import dicts from all compilers, where the key is a contract's source_id
        and the value is a list of import source_ids.

        Args:
            contract_filepaths (Sequence[pathlib.Path]): A list of source file paths to compile.
            project (Optional[:class:`~ape.managers.project.ProjectManager`]): Optionally provide
              the project.

        Returns:
            dict[str, list[str]]: A dictionary like ``{source_id: [import_source_id, ...], ...}``
        """
        imports_dict: dict[str, list[str]] = {}

        for ext, compiler in self.registered_compilers.items():
            try:
                sources = [
                    p for p in contract_filepaths if get_full_extension(p) == ext and p.is_file()
                ]
                imports = compiler.get_imports(contract_filepaths=sources, project=project)
            except NotImplementedError:
                imports = None

            if imports:
                imports_dict.update(imports)

        return imports_dict

    def get_references(self, imports_dict: dict[str, list[str]]) -> dict[str, list[str]]:
        """
        Provide a mapping containing all referenced source_ids for a given project.
        Each entry contains a source_id as a key and list of source_ids that reference a
        given contract.

        Args:
            imports_dict (dict[str, list[str]]): A dictionary of source_ids from all compilers.

        Returns:
            dict[str, list[str]]: A dictionary like ``{source_id: [referring_source_id, ...], ...}``
        """
        references_dict: dict[str, list[str]] = {}
        if not imports_dict:
            return {}

        for key, imports_list in imports_dict.items():
            for filepath in imports_list:
                if filepath not in references_dict:
                    references_dict[filepath] = []
                references_dict[filepath].append(key)

        return references_dict

    def enrich_error(self, err: ContractLogicError) -> ContractLogicError:
        """
        Enrich a contract logic error using compiler information, such
        known PC locations for compiler runtime errors.

        Args:
            err (:class:`~ape.exceptions.ContractLogicError`): The exception
              to enrich.

        Returns:
            :class:`~ape.exceptions.ContractLogicError`: The enriched exception.
        """
        # First, try enriching using their ABI.
        err = self.get_custom_error(err) or err
        if not (contract_type := err.contract_type):
            return err

        # Delegate to compiler APIs.
        elif source_id := contract_type.source_id:
            # Source ID found! Delegate to a CompilerAPI for enrichment.
            ext = get_full_extension(Path(source_id))
            if ext not in self.registered_compilers:
                # Compiler not found.
                return err

            compiler = self.registered_compilers[ext]
            return compiler.enrich_error(err)

        # No further enrichment.
        return err

    def get_custom_error(self, err: ContractLogicError) -> Optional[CustomError]:
        """
        Get a custom error for the given contract logic error using the contract-type
        found from address-data in the error. Returns ``None`` if the given error is
        not a custom-error, or it is not able to find the associated contract type or
        address.

        Args:
            err (:class:`~ape.exceptions.ContractLogicError`): The error to enrich
              as a custom error.

        Returns:
            Optional[:class:`~ape.exceptions.CustomError`]
        """
        message = err.revert_message
        if not message.startswith("0x"):
            return None
        elif not (address := err.address):
            return None

        if provider := self.network_manager.active_provider:
            ecosystem = provider.network.ecosystem
        else:
            # Default to Ethereum.
            ecosystem = self.network_manager.ethereum

        try:
            return ecosystem.decode_custom_error(
                HexBytes(message),
                address,
                base_err=err.base_err,
                source_traceback=lambda: err.source_traceback,
                trace=err.trace,
                txn=err.txn,
            )
        except NotImplementedError:
            return None

    def flatten_contract(self, path: Path, **kwargs) -> "Content":
        """
        Get the flattened version of a contract via its source path.
        Delegates to the matching :class:`~ape.api.compilers.CompilerAPI`.

        Args:
            path (``pathlib.Path``): The source path of the contract.

        Returns:
            ``ethpm_types.source.Content``: The flattened contract content.
        """

        suffix = get_full_extension(path)
        if suffix not in self.registered_compilers:
            raise CompilerError(f"Unable to flatten contract. Missing compiler for '{suffix}'.")

        compiler = self.registered_compilers[suffix]
        return compiler.flatten_contract(path, **kwargs)

    def can_trace_source(self, filename: str) -> bool:
        """
        Check if Ape is able trace the source lines for the given file.
        Checks that both the compiler is registered and that it supports
        the :meth:`~ape.api.compilers.CompilerAPI.trace_source` API method.

        Args:
            filename (str): The file to check.

        Returns:
            bool: ``True`` when the source is traceable.
        """
        path = Path(filename)
        if not path.is_file():
            return False

        extension = get_full_extension(path)
        if extension in self.registered_compilers:
            compiler = self.registered_compilers[extension]
            if compiler.supports_source_tracing:
                return True

        # We are not able to get coverage for this file.
        return False



import atexit
import os.path
import re
import shutil
from pathlib import Path
from subprocess import DEVNULL, PIPE, Popen
from typing import TYPE_CHECKING, Any, Optional, Union
from urllib.error import HTTPError, URLError
from urllib.parse import urlparse
from urllib.request import urlopen

from eth_utils import add_0x_prefix, to_hex
from geth.chain import initialize_chain as initialize_gethdev_chain
from geth.process import BaseGethProcess
from geth.wrapper import ALL_APIS, construct_test_chain_kwargs
from pydantic import field_validator
from pydantic_settings import SettingsConfigDict
from requests.exceptions import ConnectionError

from ape.api.config import PluginConfig
from ape.api.providers import SubprocessProvider, TestProviderAPI
from ape.exceptions import VirtualMachineError
from ape.logging import LogLevel, logger
from ape.utils._web3_compat import ExtraDataToPOAMiddleware
from ape.utils.misc import ZERO_ADDRESS, log_instead_of_fail, raises_not_implemented
from ape.utils.process import JoinableQueue, spawn
from ape.utils.testing import (
    DEFAULT_NUMBER_OF_TEST_ACCOUNTS,
    DEFAULT_TEST_ACCOUNT_BALANCE,
    DEFAULT_TEST_CHAIN_ID,
    DEFAULT_TEST_HD_PATH,
    DEFAULT_TEST_MNEMONIC,
    generate_dev_accounts,
)
from ape_ethereum.provider import (
    DEFAULT_HOSTNAME,
    DEFAULT_PORT,
    DEFAULT_SETTINGS,
    EthereumNodeProvider,
)
from ape_ethereum.trace import TraceApproach

if TYPE_CHECKING:
    from geth.types import GenesisDataTypedDict

    from ape.api.accounts import TestAccountAPI
    from ape.api.transactions import ReceiptAPI, TransactionAPI
    from ape.types.vm import SnapshotID


Alloc = dict[str, dict[str, Any]]


def create_genesis_data(alloc: Alloc, chain_id: int) -> "GenesisDataTypedDict":
    """
    A wrapper around genesis data for py-geth that
    fills in more defaults.
    """
    return {
        "alloc": alloc,
        "config": {
            "arrowGlacierBlock": 0,
            "berlinBlock": 0,
            "byzantiumBlock": 0,
            "cancunTime": 0,
            "chainId": chain_id,
            "constantinopleBlock": 0,
            "daoForkBlock": 0,
            "daoForkSupport": True,
            "eip150Block": 0,
            "eip155Block": 0,
            "eip158Block": 0,
            "ethash": {},
            "grayGlacierBlock": 0,
            "homesteadBlock": 0,
            "istanbulBlock": 0,
            "londonBlock": 0,
            "petersburgBlock": 0,
            "shanghaiTime": 0,
            "terminalTotalDifficulty": 0,
            "terminalTotalDifficultyPassed": True,
            "blobSchedule": {
                "cancun": {"target": 3, "max": 6, "baseFeeUpdateFraction": 3338477},
                "prague": {"target": 6, "max": 9, "baseFeeUpdateFraction": 5007716},
            },
        },
        "coinbase": ZERO_ADDRESS,
        "difficulty": "0x0",
        "gasLimit": "0x0",
        "extraData": "0x",
        "mixHash": "0x0000000000000000000000000000000000000000000000000000000000000000",
        "nonce": "0x0",
        "timestamp": "0x0",
        "parentHash": "0x0000000000000000000000000000000000000000000000000000000000000000",
        "baseFeePerGas": "0x0",
    }


def get_node_name_from_executable(executable: str) -> str:
    return executable.split(os.path.sep)[-1] or "geth"


class GethDevProcess(BaseGethProcess):
    """
    A developer-configured geth that only exists until disconnected.
    (Implementation detail of the local node provider).
    """

    def __init__(
        self,
        data_dir: Path,
        hostname: Optional[str] = None,
        port: Optional[int] = None,
        ipc_path: Optional[Path] = None,
        ws_hostname: Optional[str] = None,
        ws_port: Optional[str] = None,
        mnemonic: str = DEFAULT_TEST_MNEMONIC,
        number_of_accounts: int = DEFAULT_NUMBER_OF_TEST_ACCOUNTS,
        chain_id: int = DEFAULT_TEST_CHAIN_ID,
        initial_balance: Union[str, int] = DEFAULT_TEST_ACCOUNT_BALANCE,
        executable: Optional[Union[list[str], str]] = None,
        auto_disconnect: bool = True,
        extra_funded_accounts: Optional[list[str]] = None,
        hd_path: Optional[str] = DEFAULT_TEST_HD_PATH,
        block_time: Optional[int] = None,
        generate_accounts: bool = True,
        initialize_chain: bool = True,
        background: bool = False,
        verify_bin: bool = True,
        rpc_api: Optional[list[str]] = None,
    ):
        if isinstance(executable, str):
            # Legacy.
            executable = [executable]

        if executable:
            if verify_bin and (
                not Path(executable[0]).exists() and not shutil.which(executable[0])
            ):
                raise NodeSoftwareNotInstalledError()

        else:
            # Executable not specified: attempt to find one in $PATH.
            if shutil.which("geth"):
                executable = ["geth"]
            elif shutil.which("reth"):
                executable = ["reth", "node"]
            elif verify_bin:
                # TODO: Add support for more nodes, such as erigon.
                raise NodeSoftwareNotInstalledError()
            else:
                # This probably won't work if started, but the user knows that.
                executable = ["geth"]

        self.executable = executable or ["geth"]
        self._data_dir = data_dir
        self.is_running = False
        self._auto_disconnect = auto_disconnect
        self.background = background

        is_reth = executable[0].endswith("reth")

        if is_reth and hostname == "localhost":
            # Reth only supports IP format.
            hostname = "127.0.0.1"

        kwargs_ctor: dict = {
            "data_dir": self.data_dir,
            "geth_executable": executable[0],
            "network_id": f"{chain_id}",
        }
        if hostname is not None:
            kwargs_ctor["rpc_addr"] = hostname
        if port is not None:
            kwargs_ctor["rpc_port"] = f"{port}"
        if ws_hostname:
            kwargs_ctor["ws_enabled"] = True
            kwargs_ctor["ws_addr"] = ws_hostname
        if ws_port:
            kwargs_ctor["ws_enabled"] = True
            kwargs_ctor["ws_port"] = f"{ws_port}"
        if ipc_path is not None:
            kwargs_ctor["ipc_path"] = f"{ipc_path}"
        if not kwargs_ctor.get("ws_enabled"):
            kwargs_ctor["ws_api"] = None
            kwargs_ctor["ws_enabled"] = False
            kwargs_ctor["ws_addr"] = None
            kwargs_ctor["ws_port"] = None
        if block_time is not None and not is_reth:
            kwargs_ctor["dev_period"] = f"{block_time}"

        if rpc_api is None:
            # Reth also has MEV API support.
            rpc_api_str = f"{ALL_APIS},mev" if is_reth else ALL_APIS
        else:
            rpc_api_str = ",".join(rpc_api)

        kwargs_ctor["rpc_api"] = rpc_api_str
        geth_kwargs = construct_test_chain_kwargs(**kwargs_ctor)
        if is_reth:
            geth_kwargs.pop("max_peers", None)
            geth_kwargs.pop("network_id", None)
            geth_kwargs.pop("no_discover", None)

            # NOTE: --verbosity _is_ available in reth, but it is a different type (flag only).
            #   It isn't really needed anyway (I don't think).
            geth_kwargs.pop("verbosity", None)

            geth_kwargs.pop("password", None)

        # Ensure IPC path has correct name.
        if ipc_path_kwarg := geth_kwargs.get("ipc_path"):
            if ipc_path_kwarg.endswith("geth.ipc"):
                geth_kwargs["ipc_path"] = ipc_path_kwarg.replace(
                    "geth.ipc", f"{self.process_name}.ipc"
                )

        # Ensure a clean data-dir.
        self._clean()

        geth_kwargs["dev_mode"] = True
        hd_path = hd_path or DEFAULT_TEST_HD_PATH

        if generate_accounts:
            self._dev_accounts = generate_dev_accounts(
                mnemonic, number_of_accounts=number_of_accounts, hd_path=hd_path
            )
        else:
            self._dev_accounts = []

        if initialize_chain:
            addresses = [a.address for a in self._dev_accounts]
            addresses.extend(extra_funded_accounts or [])
            bal_dict = {"balance": str(initial_balance)}
            alloc = dict.fromkeys(addresses, bal_dict)
            genesis = create_genesis_data(alloc, chain_id)
            self._data_dir.mkdir(parents=True, exist_ok=True)
            initialize_gethdev_chain(genesis, self.data_dir)

        super().__init__(geth_kwargs)

        # Correct multi-word executable.
        idx = self.command.index(executable[0])
        self.command = self.command[:idx] + executable + self.command[idx + 1 :]

    @classmethod
    def from_uri(cls, uri: str, data_folder: Path, **kwargs):
        mnemonic = kwargs.get("mnemonic", DEFAULT_TEST_MNEMONIC)
        number_of_accounts = kwargs.get("number_of_accounts", DEFAULT_NUMBER_OF_TEST_ACCOUNTS)
        balance = kwargs.get("initial_balance", DEFAULT_TEST_ACCOUNT_BALANCE)
        extra_accounts = [a.lower() for a in kwargs.get("extra_funded_accounts", [])]
        block_time = kwargs.get("block_time", None)
        if isinstance(block_time, int):
            block_time = f"{block_time}"

        executable = kwargs.get("executable")

        process_kwargs = {
            "auto_disconnect": kwargs.get("auto_disconnect", True),
            "background": kwargs.get("background", False),
            "block_time": block_time,
            "executable": executable,
            "extra_funded_accounts": extra_accounts,
            "hd_path": kwargs.get("hd_path", DEFAULT_TEST_HD_PATH),
            "initial_balance": balance,
            "mnemonic": mnemonic,
            "number_of_accounts": number_of_accounts,
            "verify_bin": kwargs.get("verify_bin", True),
        }

        parsed_uri = urlparse(uri)
        if not parsed_uri.netloc:
            path = Path(parsed_uri.path)
            if path.suffix == ".ipc":
                # Was given an IPC path.
                process_kwargs["ipc_path"] = path

            else:
                raise ConnectionError(f"Unrecognized path type: '{path}'.")

        elif hostname := parsed_uri.hostname:
            if hostname not in ("localhost", "127.0.0.1"):
                name = get_node_name_from_executable(executable) if executable else "geth"
                raise ConnectionError(
                    f"Unable to start {name} on non-local host {parsed_uri.hostname}."
                )

            if parsed_uri.scheme.startswith("ws"):
                process_kwargs["ws_hostname"] = hostname
                process_kwargs["ws_port"] = parsed_uri.port or DEFAULT_PORT
            elif parsed_uri.scheme.startswith("http"):
                process_kwargs["hostname"] = hostname or DEFAULT_HOSTNAME
                process_kwargs["port"] = parsed_uri.port or DEFAULT_PORT
            else:
                raise ConnectionError(f"Unsupported scheme: '{parsed_uri.scheme}'.")

        return cls(data_folder, **process_kwargs)

    @property
    def data_dir(self) -> str:
        return f"{self._data_dir}"

    @property
    def process_name(self) -> str:
        return get_node_name_from_executable(self.executable[0])

    @property
    def ipc_path(self) -> str:
        # Overridden: so we can use a custom IPC path name (e.g. reth.ipc).
        return self.geth_kwargs.get("ipc_path") or os.path.abspath(
            os.path.expanduser(
                os.path.join(
                    self.data_dir,
                    f"{self.process_name}.ipc",
                )
            )
        )

    @property
    def is_rpc_ready(self) -> bool:
        # Overridden: This is overridden to work with other nodes besides `geth`.
        #   Otherwise, the RPC is never declared as ready even though it is.
        try:
            urlopen(f"http://{self.rpc_host}:{self.rpc_port}")

        except HTTPError:
            # Reth nodes (and maybe others) might throw an HTTP error here, like "method not found".
            # This means the RPC is ready.
            return True

        except URLError:
            # Nothing found at all yet, most likely.
            return False

        else:
            # No error occurs on Geth nodes when the RPC is ready.
            return True

    @property
    def _hostname(self) -> Optional[str]:
        return self.geth_kwargs.get("rpc_addr")

    @property
    def _port(self) -> Optional[str]:
        return self.geth_kwargs.get("rpc_port")

    @property
    def _ws_hostname(self) -> Optional[str]:
        return self.geth_kwargs.get("ws_addr")

    @property
    def _ws_port(self) -> Optional[str]:
        return self.geth_kwargs.get("ws_port")

    def connect(self, timeout: int = 60):
        self._log_connection()
        self.start()
        self.wait_for_rpc(timeout=timeout)

        # Register atexit handler to make sure disconnect is called for normal object lifecycle.
        if self._auto_disconnect:
            atexit.register(self.disconnect)

    def _log_connection(self):
        home = str(Path.home())
        ipc_path = self.ipc_path.replace(home, "$HOME")

        http_log = ""
        if self._hostname:
            http_log = f"HTTP={self._hostname}"
            if port := self._port:
                http_log = f"{http_log}:{port}"

        ipc_log = f"IPC={ipc_path}"

        ws_log = ""
        if self._ws_hostname:
            ws_log = f"WS={self._ws_hostname}"
            if port := self._ws_port:
                ws_log = f"{ws_log}:{port}"

        connection_logs = ", ".join(x for x in (http_log, ipc_log, ws_log) if x)
        logger.info(f"Starting {self.process_name} ({connection_logs}).")

    def start(self):
        if self.is_running:
            return

        self.is_running = True

        if self.background or logger.level > LogLevel.DEBUG:
            out_file = DEVNULL
        else:
            out_file = PIPE

        self.proc = Popen(
            self.command,
            stdin=PIPE,
            stdout=out_file,
            stderr=out_file,
        )

    def disconnect(self):
        if self.is_running:
            logger.info(f"Stopping '{self.process_name}' process.")
            self.stop()

        self._clean()

    def _clean(self):
        if self._data_dir.is_dir():
            shutil.rmtree(self._data_dir)

    def wait(self, *args, **kwargs):
        if self.proc is None:
            return

        self.proc.wait(*args, **kwargs)


class EthereumNetworkConfig(PluginConfig):
    # Make sure you are running the right networks when you try for these
    mainnet: dict = {}
    holesky: dict = {}
    sepolia: dict = {}
    # Make sure to run via `geth --dev` (or similar)
    local: dict = {**DEFAULT_SETTINGS.copy(), "chain_id": DEFAULT_TEST_CHAIN_ID, "block_time": 0}

    model_config = SettingsConfigDict(extra="allow", env_prefix="APE_NODE_")

    @field_validator("local", mode="before")
    @classmethod
    def _validate_local(cls, value):
        value = value or {}
        if not value:
            return {**DEFAULT_SETTINGS.copy(), "chain_id": DEFAULT_TEST_CHAIN_ID}

        if "chain_id" not in value:
            value["chain_id"] = DEFAULT_TEST_CHAIN_ID
        if ("uri" not in value and "ipc_path" in value) or "ws_uri" in value or "http_uri" in value:
            # No need to add default HTTP URI if was given only IPC Path
            return {**{k: v for k, v in DEFAULT_SETTINGS.items() if k != "uri"}, **value}

        return {**DEFAULT_SETTINGS, **value}


class EthereumNodeConfig(PluginConfig):
    """
    Configure your ``node:`` in Ape, the default provider
    plugin for live-network nodes. Also, ``ape node`` can
    start-up a local development node for testing purposes.
    """

    ethereum: EthereumNetworkConfig = EthereumNetworkConfig()
    """
    Configure the Ethereum network settings for the ``ape node`` provider,
    such as which URIs to use for each network.
    """

    executable: Optional[list[str]] = None
    """
    For starting nodes, select the executable. Defaults to using
    ``shutil.which("geth")``.
    """

    data_dir: Optional[Path] = None
    """
    For node-management, choose where the geth data directory shall
    be located. Defaults to using a location within Ape's DATA_FOLDER.
    """

    ipc_path: Optional[Path] = None
    """
    For IPC connections, select the IPC path. If managing a process,
    web3.py can determine the IPC w/o needing to manually configure.
    """

    call_trace_approach: Optional[TraceApproach] = None
    """
    Select the trace approach to use. Defaults to deducing one
    based on your node's client-version and available RPCs.
    """

    request_headers: dict = {}
    """
    Optionally specify request headers to use whenever using this provider.
    """

    rpc_api: Optional[list[str]] = None
    """
    RPC APIs to enable. Defaults to all geth APIs.
    """

    model_config = SettingsConfigDict(extra="allow", env_prefix="APE_NODE_")

    @field_validator("call_trace_approach", mode="before")
    @classmethod
    def validate_trace_approach(cls, value):
        # This handles nicer config values.
        return None if value is None else TraceApproach.from_key(value)

    @field_validator("executable", mode="before")
    @classmethod
    def validate_executable(cls, value):
        if not value:
            return None

        elif isinstance(value, str):
            return value.split(" ")

        return value


class NodeSoftwareNotInstalledError(ConnectionError):
    def __init__(self):
        super().__init__(
            "No node found and 'ape-node' is unable to start one.\n"
            "Things you can do:\n"
            "\t1. Check your connection URL, if trying to connect remotely.\n"
            "\t2. Install node software (geth), if trying to run a local node.\n"
            "\t3. Use and configure a different provider plugin, such as 'ape-foundry'."
        )


# NOTE: Using EthereumNodeProvider because of it's geth-derived default behavior.
# TODO: In 0.9, change NAME to be `gethdev`, so for local networks it is more obvious.
class GethDev(EthereumNodeProvider, TestProviderAPI, SubprocessProvider):
    _process: Optional[GethDevProcess] = None
    name: str = "node"

    @property
    def process_name(self) -> str:
        if self._process:
            return self._process.process_name

        elif exec_cfg := self.config.executable:
            return get_node_name_from_executable(exec_cfg[0])

        return "geth"

    @property
    def chain_id(self) -> int:
        return self.settings.ethereum.local.get("chain_id", DEFAULT_TEST_CHAIN_ID)

    @property
    def block_time(self) -> Optional[int]:
        return self.settings.ethereum.local.get("block_time")

    @property
    def data_dir(self) -> Path:
        # Overridden from base class for placing debug logs in ape data folder.
        return self.settings.data_dir or self.config_manager.DATA_FOLDER / self.name

    @log_instead_of_fail(default="<node>")
    def __repr__(self) -> str:
        client_version = self.client_version
        client_version_str = f" ({client_version}) " if client_version else " "
        return f"<Node{client_version_str}chain_id={self.chain_id}>"

    @property
    def auto_mine(self) -> bool:
        if self.process is not None:
            # Geth --dev auto mines.
            return True

        try:
            return self.make_request("eth_mining", [])
        except NotImplementedError:
            # Assume true; unlikely to be off. Geth --dev automines.
            return True

    @auto_mine.setter
    def auto_mine(self, value):
        raise NotImplementedError("'auto_mine' setter not implemented.")

    @property
    def ipc_path(self) -> Optional[Path]:
        if rpc := self._configured_ipc_path:
            # "ipc_path" found in config/settings
            return Path(rpc)

        elif rpc := self._configured_uri:
            if f"{rpc}".endswith(".ipc"):
                # "uri" found in config/settings and is IPC.
                return Path(rpc)

        elif proc := self._process:
            # Connected.
            return Path(proc.ipc_path)

        # Default (used by geth-process).
        return self.data_dir / self.process_name / f"{self.process_name}.ipc"

    def connect(self):
        self._set_web3()
        if self.is_connected:
            self._complete_connect()

        elif self.allow_start:
            # Starting the process.
            self.start()
            atexit.register(self._disconnect_atexit)

    def start(self, timeout: int = 20):
        geth_dev = self._create_process()
        geth_dev.connect(timeout=timeout)
        if not self.web3.is_connected():
            geth_dev.disconnect()

            raise ConnectionError(f"Unable to connect to locally running {geth_dev.process_name}.")
        else:
            self.web3.middleware_onion.inject(ExtraDataToPOAMiddleware, layer=0)

        self._process = geth_dev

        # For subprocess-provider
        if self._process is not None and (process := self._process.proc):
            self.stderr_queue = JoinableQueue()
            self.stdout_queue = JoinableQueue()

            self.process = process

            # Start listening to output.
            spawn(self.produce_stdout_queue)
            spawn(self.produce_stderr_queue)
            spawn(self.consume_stdout_queue)
            spawn(self.consume_stderr_queue)

            self.network_manager.running_nodes.cache_provider(self)

    def _create_process(self) -> GethDevProcess:
        # NOTE: Using JSON mode to ensure types can be passed as CLI args.
        test_config = self.config_manager.get_config("test").model_dump(mode="json")

        # Allow configuring a custom executable besides your $PATH geth.
        if self.settings.executable is not None:
            test_config["executable"] = self.settings.executable

        test_config["ipc_path"] = self.ipc_path

        # Let the provider handle disconnecting the process.
        # This avoids multiple atexit handlers from being unnecessarily
        # registered that do some of the same thing.
        test_config["auto_disconnect"] = False

        # Include extra accounts to allocated funds to at genesis.
        extra_accounts = self.settings.ethereum.local.get("extra_funded_accounts", [])
        extra_accounts.extend(self.provider_settings.get("extra_funded_accounts", []))
        extra_accounts = list({a.lower() for a in extra_accounts})
        test_config["extra_funded_accounts"] = extra_accounts
        test_config["initial_balance"] = self.test_config.balance
        test_config["background"] = self.background
        uri = self.ws_uri or self.uri
        proc_data_dir = self.data_dir / f"{self.process_name}"
        return GethDevProcess.from_uri(
            uri,
            proc_data_dir,
            block_time=self.block_time,
            **test_config,
        )

    def disconnect(self):
        # Must disconnect process first.
        if self._process is not None:
            self._process.disconnect()
            self._process = None

            # Remove self from managed-processes list.
            self.network_manager.running_nodes.remove_provider(self)

        # Also unset the subprocess-provider reference.
        # NOTE: Type ignore is wrong; TODO: figure out why.
        self.process = None  # type: ignore[assignment]

        # Clear any snapshots.
        self.chain_manager._snapshots[self.chain_id] = []

        super().disconnect()

    def send_transaction(self, txn: "TransactionAPI") -> "ReceiptAPI":
        return self._send_transaction_with_retries(txn)

    def _send_transaction_with_retries(
        self, txn: "TransactionAPI", nonce_retries: int = 0, max_nonce_retries: int = 5
    ) -> "ReceiptAPI":
        try:
            return super().send_transaction(txn)
        except VirtualMachineError as err:
            if (
                txn.sender in self.account_manager.test_accounts
                and "exceeds block gas limit" in str(err)
            ):
                # Changed, possibly due to other transactions (x-dist?).
                # Retry using block gas limit.
                block_gas_limit = self.chain_manager.blocks.head.gas_limit
                if txn.gas_limit > block_gas_limit:
                    txn.gas_limit = block_gas_limit
                elif txn.gas_limit == block_gas_limit:
                    txn.gas_limit -= 1
                else:
                    # Raise whatever error it is. I am not sure how this is possible!
                    raise

                account = self.account_manager.test_accounts[txn.sender]
                signed_transaction = account.sign_transaction(txn)
                logger.debug("Gas-limit exceeds block gas limit. Retrying using block gas limit.")
                return super().send_transaction(signed_transaction)

            elif txn.sender in self.account_manager.test_accounts and re.match(
                r".*Nonce '\d*' is too low.*", str(err)
            ):
                retries = nonce_retries + 1
                if retries > max_nonce_retries:
                    raise  # This error.

                # Try again with a new nonce.
                account = self.account_manager.test_accounts[txn.sender]
                txn.nonce = account.nonce
                signed_transaction = account.sign_transaction(txn)
                logger.debug("Test transaction received bad nonce. Retrying using latest nonce.")
                return self._send_transaction_with_retries(
                    signed_transaction,
                    nonce_retries=retries,
                    max_nonce_retries=max_nonce_retries,
                )

            raise  # Whatever error it already is (Ape-ified from ape-ethereum.provider base).

    def snapshot(self) -> "SnapshotID":
        return self._get_latest_block().number or 0

    def restore(self, snapshot_id: "SnapshotID"):
        if isinstance(snapshot_id, int):
            block_number_int = snapshot_id
            block_number_hex_str = str(to_hex(snapshot_id))
        elif isinstance(snapshot_id, bytes):
            block_number_hex_str = add_0x_prefix(to_hex(snapshot_id))
            block_number_int = int(block_number_hex_str, 16)
        else:
            block_number_hex_str = to_hex(snapshot_id)
            block_number_int = int(snapshot_id, 16)

        current_block = self._get_latest_block().number
        if block_number_int == current_block:
            # Head is already at this block.
            return
        elif block_number_int > block_number_int:
            logger.error("Unable to set head to future block.")
            return

        self.make_request("debug_setHead", [block_number_hex_str])

    @raises_not_implemented
    def set_timestamp(self, new_timestamp: int):
        pass

    @raises_not_implemented
    def mine(self, num_blocks: int = 1):
        pass

    def build_command(self) -> list[str]:
        return self._process.command if self._process else []

    def get_test_account(self, index: int) -> "TestAccountAPI":
        if self._process is None:
            # Not managing the process. Use default approach.
            test_container = self.account_manager.test_accounts.containers["test"]
            return test_container.generate_account(index)

        # perf: we avoid having to generate account keys twice by utilizing
        #   the accounts generated for geth's genesis.json.
        account = self._process._dev_accounts[index]
        return self.account_manager.init_test_account(index, account.address, account.private_key)


# NOTE: The default behavior of EthereumNodeBehavior assumes geth.
class Node(EthereumNodeProvider):
    pass



import os
import re
import shutil
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Optional, Union

import pytest
from pydantic import ValidationError
from pydantic_settings import SettingsConfigDict

from ape.api.config import ApeConfig, ConfigEnum, PluginConfig
from ape.exceptions import ConfigError
from ape.managers.config import CONFIG_FILE_NAME, merge_configs
from ape.utils.os import create_tempdir
from ape_cache.config import CacheConfig
from ape_compile.config import Config as CompileConfig
from ape_ethereum.ecosystem import EthereumConfig, ForkedNetworkConfig, NetworkConfig
from ape_networks.config import CustomNetwork
from ape_node.provider import EthereumNetworkConfig, EthereumNodeConfig
from ape_test.config import CoverageReportsConfig, GasConfig, GasExclusion
from tests.functional.conftest import PROJECT_WITH_LONG_CONTRACTS_FOLDER

if TYPE_CHECKING:
    from ape.types.gas import GasLimit


CONTRACTS_FOLDER = "pathsomewhwere"
NUMBER_OF_TEST_ACCOUNTS = 31
YAML_CONTENT = rf"""
contracts_folder: "{CONTRACTS_FOLDER}"

dependencies:
  - name: "openzeppelin"
    github: "OpenZeppelin/openzeppelin-contracts"
    version: "4.5.0"

plugins:
  - name: "hardhat"
  - name: "solidity"
    version: "0.8.1"

test:
  number_of_accounts: "{NUMBER_OF_TEST_ACCOUNTS}"

compile:
  exclude:
    - "exclude_dir"
    - "Excl*.json"
    - r"Ignore\w*\.json"
""".lstrip()
JSON_CONTENT = f"""
{{
    "contracts_folder": "{CONTRACTS_FOLDER}",
    "dependencies": [
        {{
            "name": "openzeppelin",
            "github": "OpenZeppelin/openzeppelin-contracts",
            "version": "4.5.0"
        }}
    ],
    "plugins": [
        {{
            "name": "hardhat"
        }},
        {{
            "name": "solidity",
            "version": "0.8.1"
        }}
    ],
    "test": {{
        "number_of_accounts": "{NUMBER_OF_TEST_ACCOUNTS}"
    }},
    "compile": {{
        "exclude": [
            "exclude_dir",
            "Excl*.json",
            "r\\"Ignore\\\\w*\\\\.json\\""
        ]
    }}
}}
""".lstrip()
PYPROJECT_TOML = rf"""
[tool.ape]
contracts_folder = "{CONTRACTS_FOLDER}"

[[tool.ape.dependencies]]
name = "openzeppelin"
github = "OpenZeppelin/openzeppelin-contracts"
version = "4.5.0"

[[tool.ape.plugins]]
name = "hardhat"

[[tool.ape.plugins]]
name = "solidity"
version = "0.8.1"

[tool.ape.test]
number_of_accounts = {NUMBER_OF_TEST_ACCOUNTS}

[tool.ape.compile]
exclude = ["exclude_dir", "Excl*.json", 'r"Ignore\w*\.json"']
""".lstrip()
EXT_TO_CONTENT = {
    ".yml": YAML_CONTENT,
    ".yaml": YAML_CONTENT,
    ".json": JSON_CONTENT,
    ".toml": PYPROJECT_TOML,
}


def test_model_validate_empty():
    data: dict = {}
    cfg = ApeConfig.model_validate(data)
    assert cfg.contracts_folder is None


def test_model_validate():
    data = {"contracts_folder": "src"}
    cfg = ApeConfig.model_validate(data)
    assert cfg.contracts_folder == "src"


def test_model_validate_none_contracts_folder():
    data = {"contracts_folder": None}
    cfg = ApeConfig.model_validate(data)
    assert cfg.contracts_folder is None


def test_model_validate_path_contracts_folder():
    path = Path.home() / "contracts"
    data = {"contracts_folder": path}
    cfg = ApeConfig.model_validate(data)
    assert cfg.contracts_folder == str(path)


def test_model_validate_handles_environment_variables():
    def run_test(cls: Callable, attr: str, name: str, value: str, expected: Any = None):
        expected = expected if expected is not None else value
        before: str | None = os.environ.get(name)
        os.environ[name] = value
        try:
            instance = cls()
            assert getattr(instance, attr) == expected
        finally:
            if before is not None:
                os.environ[name] = before
            else:
                os.environ.pop(name, None)

    # Test different config classes.
    run_test(ApeConfig, "contracts_folder", "APE_CONTRACTS_FOLDER", "3465220869b2")
    run_test(CacheConfig, "size", "APE_CACHE_SIZE", "8627", 8627)
    run_test(
        CompileConfig, "include_dependencies", "APE_COMPILE_INCLUDE_DEPENDENCIES", "true", True
    )
    run_test(
        ForkedNetworkConfig, "upstream_provider", "APE_ETHEREUM_UPSTREAM_PROVIDER", "411236f13659"
    )
    run_test(
        NetworkConfig, "required_confirmations", "APE_ETHEREUM_REQUIRED_CONFIRMATIONS", "6498", 6498
    )
    run_test(EthereumNetworkConfig, "mainnet", "APE_NODE_MAINNET", '{"a":"b"}', {"a": "b"})
    run_test(EthereumNodeConfig, "executable", "APE_NODE_EXECUTABLE", "geth", expected=["geth"])
    run_test(CoverageReportsConfig, "terminal", "APE_TEST_TERMINAL", "false", False)
    run_test(GasConfig, "reports", "APE_TEST_REPORTS", '["terminal"]', ["terminal"])
    run_test(GasExclusion, "method_name", "APE_TEST_METHOD_NAME", "32aa54e3c5d2")

    # Assert that union types are handled.
    run_test(NetworkConfig, "gas_limit", "APE_ETHEREUM_GAS_LIMIT", "0", 0)
    run_test(NetworkConfig, "gas_limit", "APE_ETHEREUM_GAS_LIMIT", "0x100", 0x100)
    run_test(NetworkConfig, "gas_limit", "APE_ETHEREUM_GAS_LIMIT", "auto")
    run_test(NetworkConfig, "gas_limit", "APE_ETHEREUM_GAS_LIMIT", "max")
    with pytest.raises(ValidationError, match=r"Value error, Invalid gas limit"):
        run_test(NetworkConfig, "gas_limit", "APE_ETHEREUM_GAS_LIMIT", "something")

    # Assert that various bool variants are parsed correctly.
    for bool_val in ("0", "False", "fALSE", "FALSE"):
        run_test(NetworkConfig, "is_mainnet", "APE_ETHEREUM_IS_MAINNET", bool_val, False)

    for bool_val in ("1", "True", "tRUE", "TRUE"):
        run_test(NetworkConfig, "is_mainnet", "APE_ETHEREUM_IS_MAINNET", bool_val, True)

    # We expect a failure when there's a type mismatch.
    with pytest.raises(
        ValidationError,
        match=r"Input should be a valid boolean, unable to interpret input",
    ):
        run_test(NetworkConfig, "is_mainnet", "APE_ETHEREUM_IS_MAINNET", "not a boolean", False)

    with pytest.raises(
        ValidationError,
        match=r"Input should be a valid integer, unable to parse string as an integer",
    ):
        run_test(
            NetworkConfig,
            "required_confirmations",
            "APE_ETHEREUM_REQUIRED_CONFIRMATIONS",
            "not a number",
            42,
        )


@pytest.mark.parametrize(
    "file", ("ape-config.yml", "ape-config.yaml", "ape-config.json", "pyproject.toml")
)
def test_validate_file(file):
    content = EXT_TO_CONTENT[Path(file).suffix]
    with create_tempdir() as temp_dir:
        path = temp_dir / file
        path.write_text(content)
        actual = ApeConfig.validate_file(path)

    assert actual.contracts_folder == CONTRACTS_FOLDER
    assert actual.test.number_of_accounts == NUMBER_OF_TEST_ACCOUNTS
    assert len(actual.dependencies) == 1
    assert actual.dependencies[0]["name"] == "openzeppelin"
    assert actual.dependencies[0]["github"] == "OpenZeppelin/openzeppelin-contracts"
    assert actual.dependencies[0]["version"] == "4.5.0"
    assert actual.plugins == [{"name": "hardhat"}, {"name": "solidity", "version": "0.8.1"}]
    assert re.compile("Ignore\\w*\\.json") in actual.compile.exclude
    assert "exclude_dir" in actual.compile.exclude
    assert ".cache" in actual.compile.exclude
    assert "Excl*.json" in actual.compile.exclude


def test_validate_file_expands_environment_variables():
    secret = "mycontractssecretfolder"
    env_var_name = "APE_TEST_CONFIG_SECRET_CONTRACTS_FOLDER"
    os.environ[env_var_name] = secret

    try:
        with create_tempdir() as temp_dir:
            file = temp_dir / "ape-config.yaml"
            file.write_text(f"contracts_folder: ${env_var_name}")

            actual = ApeConfig.validate_file(file)
            assert actual.contracts_folder == secret
    finally:
        if env_var_name in os.environ:
            del os.environ[env_var_name]


def test_validate_file_shows_linenos():
    with create_tempdir() as temp_dir:
        file = temp_dir / "ape-config.yaml"
        file.write_text("name: {'test': 123}")

        expected = (
            f"'{temp_dir / 'ape-config.yaml'}' is invalid!"
            "\nInput should be a valid string\n-->1: name: {'test': 123}"
        )
        with pytest.raises(ConfigError) as err:
            _ = ApeConfig.validate_file(file)

        assert expected in str(err.value)


def test_validate_file_shows_linenos_handles_lists():
    with create_tempdir() as temp_dir:
        file = temp_dir / "ape-config.yaml"
        file.write_text("deployments:\n  ethereum:\n   sepolia:\n      - foo: bar")
        with pytest.raises(ConfigError) as err:
            _ = ApeConfig.validate_file(file)

        assert str(file) in str(err.value)
        assert "sepolia:" in str(err.value)
        assert "-->4" in str(err.value)


def test_validate_file_uses_project_name():
    name = "apexampledapp"
    with create_tempdir() as temp_dir:
        file = temp_dir / "pyproject.toml"
        content = f'[project]\nname = "{name}"\n'
        file.write_text(content)
        cfg = ApeConfig.validate_file(file)
        assert cfg.name == name


def test_deployments(networks_connected_to_tester, owner, vyper_contract_container, project):
    _ = networks_connected_to_tester  # Connection needs to lookup config.

    # First, obtain a "previously-deployed" contract.
    instance = vyper_contract_container.deploy(1000200000, sender=owner)
    address = instance.address

    # Create a config using this new contract for a "later time".
    deploys = {
        **_create_deployments(address=address, contract_name=instance.contract_type.name),
    }
    with project.temp_config(deployments=deploys):
        deploy_config = project.config.deployments
        assert deploy_config["ethereum"]["local"][0]["address"] == address
        deployment = vyper_contract_container.deployments[0]

    assert deployment.address == instance.address


def test_deployments_integer_type_addresses(networks, project):
    deploys = {
        **_create_deployments(address=0x0C25212C557D00024B7CA3DF3238683A35541354),
    }
    with project.temp_config(deployments=deploys):
        deploy_config = project.config.deployments
        assert (
            deploy_config["ethereum"]["local"][0]["address"]
            == "0x0c25212c557d00024b7Ca3df3238683A35541354"
        )


def test_deployments_bad_ecosystem(project):
    deployments = _create_deployments(ecosystem_name="madeup")
    with project.temp_config(deployments=deployments):
        with pytest.raises(
            ConfigError, match=r"Invalid ecosystem 'madeup' in deployments config\."
        ):
            _ = project.config.deployments


def test_deployments_bad_network(project):
    deployments = _create_deployments(network_name="madeup")
    with project.temp_config(deployments=deployments):
        with pytest.raises(
            ConfigError, match=r"Invalid network 'ethereum:madeup' in deployments config\."
        ):
            _ = project.config.deployments


def _create_deployments(
    ecosystem_name: str = "ethereum",
    network_name: str = "local",
    address: Union[int, str] = "0x0C25212C557D00024B7CA3DF3238683A35541354",
    contract_name: Optional[str] = "MyContract",
) -> dict:
    return {
        ecosystem_name: {
            network_name: [
                {
                    "address": address,
                    "contract_type": contract_name,
                }
            ]
        }
    }


def test_ethereum_network_configs(config, project):
    eth_config = {"ethereum": {"sepolia": {"default_provider": "test"}}}
    with project.temp_config(**eth_config):
        actual = config.get_config("ethereum")
        assert actual.sepolia.default_provider == "test"

        # Ensure that non-updated fields remain unaffected
        assert actual.sepolia.block_time == 15


def test_network_gas_limit_default(config):
    eth_config = config.get_config("ethereum")

    assert eth_config.sepolia.gas_limit == "auto"
    assert eth_config.local.gas_limit == "max"


def _sepolia_with_gas_limit(gas_limit: "GasLimit") -> dict:
    return {
        "ethereum": {
            "sepolia": {
                "default_provider": "test",
                "gas_limit": gas_limit,
            }
        }
    }


@pytest.mark.parametrize("gas_limit", ("auto", "max"))
def test_network_gas_limit_string_config(gas_limit, project):
    eth_config = _sepolia_with_gas_limit(gas_limit)

    with project.temp_config(**eth_config):
        actual = project.config.get_config("ethereum")

        assert actual.sepolia.gas_limit == gas_limit

        # Local configuration is unaffected
        assert actual.local.gas_limit == "max"


@pytest.mark.parametrize("gas_limit", (1234, "1234", 0x4D2, "0x4D2"))
def test_network_gas_limit_numeric_config(gas_limit, project):
    eth_config = _sepolia_with_gas_limit(gas_limit)
    with project.temp_config(**eth_config):
        actual = project.config.get_config("ethereum")
        assert actual.sepolia.gas_limit == 1234

        # Local configuration is unaffected
        assert actual.local.gas_limit == "max"


def test_network_gas_limit_invalid_numeric_string(project):
    """
    Test that using hex strings for a network's gas_limit config must be
    prefixed with '0x'
    """
    sep_cfg = _sepolia_with_gas_limit("4D2")["ethereum"]["sepolia"]
    with pytest.raises(ValidationError, match="Gas limit hex str must include '0x' prefix."):
        NetworkConfig.model_validate(sep_cfg)


def test_dependencies(project_with_dependency_config):
    config = project_with_dependency_config.config
    assert len(config.dependencies) == 1
    assert config.dependencies[0]["name"] == "testdependency"
    assert config.dependencies[0]["config_override"]["contracts_folder"] == "source/v0.1"
    assert config.dependencies[0]["local"] == str(PROJECT_WITH_LONG_CONTRACTS_FOLDER)


def test_config_access():
    config = NetworkConfig()
    assert "default_provider" in config
    assert (
        config.default_provider
        == config["default_provider"]
        == getattr(config, "default-provider")
        == "node"
    )


def test_from_overrides():
    class MyConfig(PluginConfig):
        foo: int = 0

    actual = MyConfig.from_overrides({"foo": 1})
    assert actual.foo == 1


def test_from_overrides_updates_when_default_is_empty_dict():
    class SubConfig(PluginConfig):
        foo: int = 0
        bar: int = 1

    class MyConfig(PluginConfig):
        sub: dict[str, dict[str, SubConfig]] = {}

    overrides = {"sub": {"baz": {"test": {"foo": 5}}}}
    actual = MyConfig.from_overrides(overrides)
    assert actual.sub == {"baz": {"test": SubConfig(foo=5, bar=1)}}


def test_from_overrides_shows_errors_in_project_config():
    class MyConfig(PluginConfig):
        foo: int = 0

    with create_tempdir() as tmp_path:
        file = tmp_path / "ape-config.yaml"
        file.write_text("foo: [1,2,3]")

        with pytest.raises(ConfigError) as err:
            _ = MyConfig.from_overrides({"foo": [1, 2, 3]}, project_path=tmp_path)

        assert "-->1: foo: [1,2,3]" in str(err.value)


@pytest.mark.parametrize("override_0,override_1", [(True, {"foo": 0}), ({"foo": 0}, True)])
def test_plugin_config_with_union_dicts(override_0, override_1):
    class SubConfig(PluginConfig):
        bool_or_dict: Union[bool, dict] = True
        dict_or_bool: Union[dict, bool] = {}

    config = SubConfig.from_overrides({"bool_or_dict": override_0, "dict_or_bool": override_1})
    assert config.bool_or_dict == override_0
    assert config.dict_or_bool == override_1


def test_global_config(data_folder, config):
    config_file = data_folder / CONFIG_FILE_NAME
    config_file.unlink(missing_ok=True)
    config_file.touch()
    config_content = """
test:
  number_of_accounts: 11
""".strip()
    config_file.write_text(config_content, encoding="utf8")
    global_config = config.load_global_config()
    assert global_config.get_config("test").number_of_accounts == 11
    config_file.unlink(missing_ok=True)


def test_merge_configs():
    """
    The test covers most cases in `merge_config()`. See comment below explaining
    `expected`.
    """
    global_config = {
        "ethereum": {
            "mainnet": {"default_provider": "node"},
            "local": {"default_provider": "test", "required_confirmations": 5},
        }
    }
    project_config = {
        "ethereum": {
            "local": {"default_provider": "node"},
            "sepolia": {"default_provider": "alchemy"},
        },
        "test": "foo",
    }
    actual = merge_configs(global_config, project_config)

    # Expected case `key only in global`: "mainnet"
    # Expected case `non-primitive override from project`: local -> default_provider, which
    #  is `test` in global but `geth` in project; thus it is `geth` in expected.
    # Expected case `merge sub-dictionaries`: `ethereum` is being merged.
    # Expected case `add missing project keys`: `test` is added, so is `sepolia` (nested-example).
    expected = {
        "ethereum": {
            "local": {"default_provider": "node", "required_confirmations": 5},
            "mainnet": {"default_provider": "node"},
            "sepolia": {"default_provider": "alchemy"},
        },
        "test": "foo",
    }
    assert actual == expected


def test_merge_configs_short_circuits():
    """
    Cover all short-circuit cases.
    """
    ex = {"test": "foo"}
    assert merge_configs({}, {}) == {}
    assert merge_configs(ex, {}) == merge_configs({}, ex) == ex


def test_merge_configs_wrong_type():
    cfg_0 = {"foo": 123}
    cfg_1 = {"foo": {"bar": 123}}
    actual = merge_configs(cfg_0, cfg_1)
    assert actual["foo"] == {"bar": 123}
    actual = merge_configs(cfg_1, cfg_0)
    assert actual["foo"] == 123


def test_plugin_config_getattr_and_getitem(config):
    config = config.get_config("ethereum")
    assert config.mainnet is not None
    assert config.mainnet == config["mainnet"]


def test_custom_plugin_config_extras():
    class CustomConfig(PluginConfig):
        model_config = SettingsConfigDict(extra="allow")

    config = CustomConfig(foo="123")
    assert "foo" in config
    assert config.foo == "123"
    assert config["foo"] == "123"


def test_config_enum():
    class MyEnum(ConfigEnum):
        FOO = "FOO"
        BAR = "BAR"

    class MyConfig(PluginConfig):
        my_enum: MyEnum

    actual = MyConfig(my_enum="FOO")
    assert actual.my_enum == MyEnum.FOO


def test_contracts_folder(project):
    with project.temp_config(**{"contracts_folder": "src"}):
        assert project.contracts_folder.name == "src"


def test_contracts_folder_with_hyphen(project):
    with project.temp_config(**{"contracts-folder": "src"}):
        assert project.contracts_folder.name == "src"


def test_contracts_folder_invalid(project):
    """
    Show that we can still attempt to use `.get_config()` when config is invalid.
    """
    with pytest.raises(ConfigError):
        with project.temp_config(**{"contracts_folder": {}}):
            _ = project.contracts_folder

    # Ensure config is fine _after_ something invalid.
    assert project.contracts_folder


def test_custom_network():
    chain_id = 11191919191991918223773
    data = {
        "name": "mytestnet",
        "chain_id": 11191919191991918223773,
        "ecosystem": "ethereum",
    }
    network = CustomNetwork.model_validate(data)
    assert network.name == "mytestnet"
    assert network.chain_id == chain_id
    assert network.ecosystem == "ethereum"


def test_get_config(config):
    actual = config.get_config("ethereum")
    assert isinstance(actual, EthereumConfig)


def test_get_config_hyphen_in_plugin_name(config):
    """
    Tests against a bug noticed with ape-polygon-zkevm installed
    on Ape 0.8.0 release where the config no longer worked.
    """

    class CustomConfig(PluginConfig):
        x: int = 123

    mock_cfg_with_hyphens = CustomConfig
    original_method = config.local_project.config._get_config_plugin_classes

    # Hack in the fake plugin to test the behavior.
    def hacked_in_method():
        yield from [*list(original_method()), ("mock-plugin", mock_cfg_with_hyphens)]

    config.local_project.config._get_config_plugin_classes = hacked_in_method

    try:
        cfg = config.get_config("mock-plugin")
        assert isinstance(cfg, CustomConfig)
        assert cfg.x == 123

    finally:
        config.local_project.config._get_config_plugin_classes = original_method


def test_get_config_unknown_plugin(config):
    """
    Simulating reading plugin configs w/o those plugins installed.
    """
    actual = config.get_config("thisshouldnotbeinstalled")
    assert isinstance(actual, PluginConfig)


def test_get_config_invalid_plugin_config(project):
    with project.temp_config(node={"ethereum": [1, 2]}):
        # Show project's ApeConfig model works.
        with pytest.raises(ConfigError):
            project.config.get_config("node")

        # Show the manager-wrapper also works
        # (simple wrapper for local project's config,
        # but at one time pointlessly overrode the `get_config()`
        # which caused issues).
        with pytest.raises(ConfigError):
            project.config_manager.get_config("node")


def test_write_to_disk_json(config):
    with create_tempdir() as base_path:
        path = base_path / "config.json"
        config.write_to_disk(path)
        assert path.is_file()


def test_write_to_disk_yaml(config):
    with create_tempdir() as base_path:
        path = base_path / "config.yaml"
        config.write_to_disk(path)
        assert path.is_file()


def test_write_to_disk_txt(config):
    with create_tempdir() as base_path:
        path = base_path / "config.txt"
        with pytest.raises(ConfigError, match=f"Unsupported destination file type '{path}'."):
            config.write_to_disk(path)


def test_dependencies_not_list_of_dicts(project):
    # NOTE: `project:` is a not a user-facing config, only
    #   for internal Ape.
    data = {"dependencies": 123, "project": str(project.path)}
    expected = "Expecting dependencies: to be iterable. Received: int"
    with pytest.raises(ConfigError, match=expected):
        _ = ApeConfig.model_validate(data)


def test_dependencies_list_of_non_dicts(project):
    # NOTE: `project:` is a not a user-facing config, only
    #   for internal Ape.
    data = {"dependencies": [123, 123], "project": str(project.path)}
    expected = "Expecting mapping for dependency. Received: int."
    with pytest.raises(ConfigError, match=expected):
        _ = ApeConfig.model_validate(data)


def test_project_level_settings(project):
    """
    Settings can be configured in an ape-config.yaml file
    that are not part of any plugin. This test ensures that
    works.
    """
    # NOTE: Using strings for the values to show simple validation occurs.
    with project.temp_config(my_string="my_string", my_int=123, my_bool=True):
        assert project.config.my_string == "my_string"
        assert project.config.my_int == 123
        assert project.config.my_bool is True


def test_isolate_data_folder(config):
    original_data_folder = config.DATA_FOLDER
    madeup_path = Path.home() / ".ape" / "__madeup__"
    config.DATA_FOLDER = madeup_path
    try:
        with config.isolate_data_folder():
            assert original_data_folder != config.DATA_FOLDER
    finally:
        config.DATA_FOLDER = original_data_folder


def test_isolate_data_folder_already_isolated(config):
    data_folder = config.DATA_FOLDER
    with config.isolate_data_folder():
        # Already isolated, so there is no change.
        assert data_folder == config.DATA_FOLDER


def test_isolate_data_folder_keep(config):
    original_data_folder = config.DATA_FOLDER
    madeup_path = Path.home() / ".ape" / "__core_aoe_test__"
    madeup_path.mkdir(parents=True, exist_ok=True)
    sub_dir = madeup_path / "subdir"
    file = sub_dir / "file.txt"
    sub_dir.mkdir(parents=True, exist_ok=True)
    file.write_text("search for 'test_isolate_data_folder_keep' ape's tests.")

    config.DATA_FOLDER = madeup_path
    try:
        with config.isolate_data_folder(keep="subdir"):
            assert original_data_folder != config.DATA_FOLDER

        expected_file = config.DATA_FOLDER / "subdir" / "file.txt"
        assert expected_file.is_file()

    finally:
        config.DATA_FOLDER = original_data_folder
        if madeup_path.is_dir():
            shutil.rmtree(str(madeup_path), ignore_errors=True)



from ape_ethereum.transactions import TransactionType
from tests.conftest import geth_process_test


@geth_process_test
def test_default_transaction_type_configured_from_custom_network(
    custom_network_connection, ethereum
):
    assert ethereum.default_transaction_type == TransactionType.STATIC



from collections.abc import Iterator
from functools import singledispatchmethod
from typing import Optional

from ape.api.query import ContractCreation, ContractCreationQuery, QueryAPI, QueryType
from ape.exceptions import APINotImplementedError, ProviderError, QueryEngineError
from ape.types.address import AddressType


class EthereumQueryProvider(QueryAPI):
    """
    Implements more advanced queries specific to Ethereum clients.
    """

    def __init__(self):
        self.supports_contract_creation = None  # will be set after we try for the first time

    @singledispatchmethod
    def estimate_query(self, query: QueryType) -> Optional[int]:  # type: ignore[override]
        return None

    @singledispatchmethod
    def perform_query(self, query: QueryType) -> Iterator:  # type: ignore[override]
        raise QueryEngineError(f"Cannot handle '{type(query)}'.")

    @estimate_query.register
    def estimate_contract_creation_query(self, query: ContractCreationQuery) -> Optional[int]:
        # NOTE: Extremely expensive query, involves binary search of all blocks in a chain
        #       Very loose estimate of 5s per transaction for this query.
        if self.supports_contract_creation is False:
            return None
        return 5000

    @perform_query.register
    def perform_contract_creation_query(
        self, query: ContractCreationQuery
    ) -> Iterator[ContractCreation]:
        """
        Find when a contract was deployed using binary search and block tracing.
        """
        # skip the search if there is still no code at address at head
        if not self.chain_manager.get_code(query.contract):
            return None

        def find_creation_block(lo, hi):
            # perform a binary search to find the block when the contract was deployed.
            # takes log2(height), doesn't work with contracts that have been reinit.
            while hi - lo > 1:
                mid = (lo + hi) // 2
                code = self.chain_manager.get_code(query.contract, block_id=mid)
                if not code:
                    lo = mid
                else:
                    hi = mid

            if self.chain_manager.get_code(query.contract, block_id=hi):
                return hi

            return None

        try:
            block = find_creation_block(0, self.chain_manager.blocks.height)
        except ProviderError:
            self.supports_contract_creation = False
            return None

        # iterate over block transaction traces to find the deployment call
        # this method also supports contracts created by factories
        try:
            if "geth" in self.provider.client_version.lower():
                yield from self._find_creation_in_block_via_geth(block, query.contract)
            else:
                yield from self._find_creation_in_block_via_parity(block, query.contract)
        except (ProviderError, APINotImplementedError):
            self.supports_contract_creation = False
            return None

        self.supports_contract_creation = True

    def _find_creation_in_block_via_parity(self, block, contract_address):
        # NOTE requires `trace_` namespace
        traces = self.provider.make_request("trace_replayBlockTransactions", [block, ["trace"]])

        for tx in traces:
            for trace in tx["trace"]:
                if (
                    "error" not in trace
                    and trace["type"] == "create"
                    and trace["result"]["address"] == contract_address.lower()
                ):
                    receipt = self.chain_manager.get_receipt(tx["transactionHash"])
                    creator = self.conversion_manager.convert(trace["action"]["from"], AddressType)
                    yield ContractCreation(
                        txn_hash=tx["transactionHash"],
                        block=block,
                        deployer=receipt.sender,
                        factory=creator if creator != receipt.sender else None,
                    )

    def _find_creation_in_block_via_geth(self, block, contract_address):
        # NOTE requires `debug_` namespace
        traces = self.provider.make_request(
            "debug_traceBlockByNumber", [hex(block), {"tracer": "callTracer"}]
        )

        def flatten(call):
            if call["type"] in ["CREATE", "CREATE2"]:
                yield call["from"], call["to"]

            if "error" in call or "calls" not in call:
                return

            for sub in call["calls"]:
                if sub["type"] in ["CREATE", "CREATE2"]:
                    yield sub["from"], sub["to"]
                else:
                    yield from flatten(sub)

        for tx in traces:
            call = tx["result"]
            sender = call["from"]
            for factory, contract in flatten(call):
                if contract == contract_address.lower():
                    yield ContractCreation(
                        txn_hash=tx["txHash"],
                        block=block,
                        deployer=self.conversion_manager.convert(sender, AddressType),
                        factory=(
                            self.conversion_manager.convert(factory, AddressType)
                            if factory != sender
                            else None
                        ),
                    )



import pytest

import ape

INITIAL_BALANCE = 1_000_1 * 10**18


@pytest.fixture(scope="function")
def function_one(chain):
    chain.mine(1)


TOKEN_MAP = {
    "WETH": "weth-token",
    "DAI": "dai-token",
    "BAT": "bat-token",
}


@ape.fixture(scope="module", chain_isolation=False, params=("WETH", "DAI", "BAT"))
def token_key(request):
    return TOKEN_MAP[request.param]


def test_token_key(token_key):
    # TODO: Improve this test - show token key doesn't trigger
    #   resets in those unfortunate conditions.
    assert True


@pytest.fixture(scope="module", autouse=True)
def module_one(chain):
    chain.mine(3)


@pytest.fixture(scope="session")
def alice(accounts):
    return accounts[0]


@pytest.fixture(scope="session")
def bob(accounts):
    return accounts[1]


@pytest.fixture(params=(5, 6, 7))
def parametrized_mining(chain, request):
    chain.mine(request.param)
    return request.param


@pytest.fixture(scope="module", autouse=True)
def setup(alice, bob, chain):
    start_number = chain.provider.get_block("latest").number
    alice.transfer(bob, 10**18)
    actual = chain.provider.get_block("latest").number
    expected = start_number + 1
    assert actual == expected


@pytest.fixture(scope="module")
def start_block_number(chain):
    return chain.blocks.height


def test_noop():
    # This forces auto-use fixtures to fire off before
    # any requested fixtures, for testing purposes.
    assert True


class TestClass:
    @pytest.fixture(scope="class", autouse=True)
    def classminer(self, chain):
        chain.mine(9)

    def test_chain(self, chain):
        """
        Sessions haven't run yet.
        Module mined 3 + 1 = 4 total.
        Class mined 9.
        4 + 9 = 13
        """
        assert chain.blocks.height == 13


def test_isolation_first(alice, bob, chain, start_block_number):
    assert chain.provider.get_block("latest").number == start_block_number
    assert bob.balance == INITIAL_BALANCE
    alice.transfer(bob, "1 ether")


def test_isolation_second(bob, chain, start_block_number):
    assert chain.provider.get_block("latest").number == start_block_number
    assert bob.balance == INITIAL_BALANCE


def test_isolation_with_session_module_and_function(chain, session_one, session_two, function_one):
    """
    The sessions should be used, so that is 6.
    Function is 1 and the module 3.
    Also, setup does a transfer - that bumps up another 1.
    Expected is 11.
    """
    # NOTE: Module is on autouse=True
    assert chain.blocks.height == 11


def test_isolation_module_ran_after(chain):
    """
    This test runs after the test above.
    We should be back at the beginning of the state after
    the session and module function but before the function.
    Expected = sessions + module = 4 + 2 + 3 + 1 (from setup) = 10
    """
    assert chain.blocks.height == 10


def test_parametrized_fixtures(start_block_number, chain, parametrized_mining):
    assert chain.blocks.height == start_block_number + parametrized_mining


@pytest.fixture(scope="session", params=(1, 2, 3))
def parametrized_transaction(request, alice, bob):
    """
    3 more get added to the session here!
    """
    alice.transfer(bob, f"{request.param} wei")
    return request.param


@pytest.fixture(scope="session", params=(1, 2, 3))
def second_parametrized_transaction(request, alice, bob):
    """
    2 more get added to the session here!
    """
    alice.transfer(bob, f"{request.param * 2} wei")
    return request.param


@pytest.fixture
def functional_fixture_using_session(chain, session_one):
    """
    Showing the transactions in a functional-scoped
    fixture that use a session-scoped fixture don't
    persist on-chain.
    """
    _ = session_one
    chain.mine()
    return 11  # expected: 10 built up plus this 1.


# Parametrized to show it works more than once.
@pytest.mark.parametrize("it", (0, 1, 2))
def test_functional_fixture_using_session(chain, functional_fixture_using_session, it):
    assert chain.blocks.height == functional_fixture_using_session


def test_use_parametrized_transaction(chain, parametrized_transaction):
    starting = 10  # All session + module
    assert chain.blocks.height == starting + parametrized_transaction


def test_use_parametrized_transaction_again(chain, parametrized_transaction):
    """
    Should not have invalidated parametrized fixture.
    """
    starting = 10  # All session + module
    assert chain.blocks.height == starting + parametrized_transaction


@pytest.fixture
def functional_fixture_using_parametrized_session(chain, parametrized_transaction):
    chain.mine()
    return 11 + parametrized_transaction


def test_functional_fixture_using_parametrized_session(
    chain, functional_fixture_using_parametrized_session
):
    assert chain.blocks.height == functional_fixture_using_parametrized_session


@pytest.mark.parametrize("foo", (1, 2, 3))
def test_parametrized_test(foo):
    """
    Ensuring parametrized tests don't mess up our isolation-fixture logic
    (it was the case at one point!)
    """
    assert isinstance(foo, int)


def test_use_isolate_in_test(chain, parametrized_transaction):
    """
    Show the isolation we control doesn't affect
    the isolation fixtures.
    """
    _ = parametrized_transaction  # Using this for complexity.
    start_block = chain.blocks.height
    with chain.isolate():
        chain.mine()
        assert chain.blocks.height == start_block + 1

    assert chain.blocks.height == start_block



import contextlib
from collections.abc import Generator, Iterator
from contextlib import AbstractContextManager as ContextManager
from functools import cached_property, singledispatchmethod
from typing import TYPE_CHECKING, Optional, Union

from eth_utils import is_hex

from ape.api import (
    AccountAPI,
    AccountContainerAPI,
    ImpersonatedAccount,
    TestAccountAPI,
    TestAccountContainerAPI,
)
from ape.exceptions import AccountsError, ConversionError
from ape.managers.base import BaseManager
from ape.types.address import AddressType
from ape.utils.basemodel import ManagerAccessMixin
from ape.utils.misc import log_instead_of_fail

if TYPE_CHECKING:
    from ape.api.address import BaseAddress

_DEFAULT_SENDERS: list[AccountAPI] = []


@contextlib.contextmanager
def _use_sender(
    account: Union[AccountAPI, TestAccountAPI],
) -> "Generator[AccountAPI, TestAccountAPI, None]":
    try:
        _DEFAULT_SENDERS.append(account)
        yield account
    finally:
        _DEFAULT_SENDERS.pop()


class TestAccountManager(list, ManagerAccessMixin):
    __test__ = False

    _impersonated_accounts: dict[AddressType, ImpersonatedAccount] = {}
    _accounts_by_index: dict[int, AccountAPI] = {}

    @log_instead_of_fail(default="<TestAccountManager>")
    def __repr__(self) -> str:
        return f"<apetest-wallet {self.hd_path}>"

    @cached_property
    def containers(self) -> dict[str, TestAccountContainerAPI]:
        account_types = filter(
            lambda t: issubclass(t[1][1], TestAccountAPI), self.plugin_manager.account_types
        )
        return {
            plugin_name: container_type(name=plugin_name, account_type=account_type)
            for plugin_name, (container_type, account_type) in account_types
        }

    @property
    def mnemonic(self) -> str:
        """
        The seed phrase for generated test accounts.
        """
        return self.config_manager.get_config("test").mnemonic

    @mnemonic.setter
    def mnemonic(self, value: str):
        """
        The seed phrase for generated test accounts.
        **WARNING**: Changing the test-mnemonic mid-session
           re-starts the provider (if connected to one).
        """
        self.config_manager.test.mnemonic = value
        self.containers["test"].mnemonic = value

        if provider := self.network_manager.active_provider:
            provider.update_settings({"mnemonic": value})

        self._accounts_by_index = {}

    @property
    def number_of_accounts(self) -> int:
        """
        The number of test accounts to generate and fund by default.
        """
        return self.config_manager.test.number_of_accounts

    @property
    def hd_path(self) -> str:
        """
        The HD path used for generating the test accounts.
        """
        return self.config_manager.get_config("test").hd_path

    @property
    def accounts(self) -> Iterator[AccountAPI]:
        for container in self.containers.values():
            yield from container.accounts

    def aliases(self) -> Iterator[str]:
        for account in self.accounts:
            if account.alias:
                yield account.alias

    def __len__(self) -> int:
        return sum(len(c) for c in self.containers.values())

    def __iter__(self) -> Iterator[AccountAPI]:
        yield from self.accounts

    @singledispatchmethod
    def __getitem__(self, account_id):
        raise NotImplementedError(f"Cannot use {type(account_id)} as account ID.")

    @__getitem__.register
    def __getitem_int(self, account_id: int):
        if account_id < 0:
            account_id = len(self) + account_id

        if account_id in self._accounts_by_index:
            return self._accounts_by_index[account_id]

        account = self.containers["test"].get_test_account(account_id)
        self._accounts_by_index[account_id] = account
        return account

    @__getitem__.register
    def __getitem_slice(self, account_id: slice):
        start_idx = account_id.start or 0
        if start_idx < 0:
            start_idx += len(self)
        stop_idx = account_id.stop or len(self)
        if stop_idx < 0:
            stop_idx += len(self)
        step_size = account_id.step or 1
        return [self[i] for i in range(start_idx, stop_idx, step_size)]

    @__getitem__.register
    def __getitem_str(self, account_str: str):
        message_fmt = "No account with {} '{}'."
        try:
            account_id = self.conversion_manager.convert(account_str, AddressType)
        except ConversionError as err:
            message = message_fmt.format("ID", account_str)
            raise KeyError(message) from err

        for account in self.accounts:
            if account.address == account_id:
                return account

        try:
            return self.impersonate_account(account_id)
        except AccountsError as err:
            err_message = message_fmt.format("address", account_id)
            raise KeyError(f"{str(err).rstrip('.')}:\n{err_message}") from err

    def __contains__(self, address: AddressType) -> bool:  # type: ignore
        return any(address in container for container in self.containers.values())

    def impersonate_account(self, address: AddressType) -> ImpersonatedAccount:
        """
        Impersonate an account for testing purposes.

        Args:
            address (AddressType): The address to impersonate.
        """
        try:
            result = self.provider.unlock_account(address)
        except NotImplementedError as err:
            raise AccountsError(
                f"Provider '{self.provider.name}' does not support impersonating accounts."
            ) from err

        if result:
            if address in self._impersonated_accounts:
                return self._impersonated_accounts[address]

            account = ImpersonatedAccount(raw_address=address)
            self._impersonated_accounts[address] = account
            return account

        raise AccountsError(f"Unable to unlocked account '{address}'.")

    def stop_impersonating(self, address: AddressType):
        """
        End the impersonating of an account, if it is being impersonated.

        Args:
            address (AddressType): The address to stop impersonating.
        """
        if address in self._impersonated_accounts:
            del self._impersonated_accounts[address]

        try:
            self.provider.relock_account(address)
        except NotImplementedError:
            pass

    def generate_test_account(self, container_name: str = "test") -> TestAccountAPI:
        return self.containers[container_name].generate_account()

    def use_sender(self, account_id: Union[TestAccountAPI, AddressType, int]) -> "ContextManager":
        account = account_id if isinstance(account_id, TestAccountAPI) else self[account_id]
        return _use_sender(account)

    def init_test_account(
        self, index: int, address: AddressType, private_key: str
    ) -> "TestAccountAPI":
        container = self.containers["test"]
        return container.init_test_account(  # type: ignore[attr-defined]
            index, address, private_key
        )

    def reset(self):
        self._accounts_by_index = {}
        for container in self.containers.values():
            container.reset()


class AccountManager(BaseManager):
    """
    The ``AccountManager`` is a container of containers for
    :class:`~ape.api.accounts.AccountAPI` objects.
    All containers must subclass :class:`~ape.api.accounts.AccountContainerAPI`
    and are treated as singletons.

    Import the accounts manager singleton from the root ``ape`` namespace.

    Usage example::

        from ape import accounts  # "accounts" is the AccountManager singleton

        my_accounts = accounts.load("dev")
    """

    _alias_to_account_cache: dict[str, AccountAPI] = {}

    @property
    def default_sender(self) -> Optional[AccountAPI]:
        return _DEFAULT_SENDERS[-1] if _DEFAULT_SENDERS else None

    @cached_property
    def containers(self) -> dict[str, AccountContainerAPI]:
        """
        A dict of all :class:`~ape.api.accounts.AccountContainerAPI` instances
        across all installed plugins.

        Returns:
            dict[str, :class:`~ape.api.accounts.AccountContainerAPI`]
        """
        containers = {}
        data_folder = self.config_manager.DATA_FOLDER
        data_folder.mkdir(exist_ok=True)
        for plugin_name, (container_type, account_type) in self.plugin_manager.account_types:
            # Ignore containers that contain test accounts.
            if issubclass(account_type, TestAccountAPI):
                continue

            containers[plugin_name] = container_type(name=plugin_name, account_type=account_type)

        return containers

    @property
    def aliases(self) -> Iterator[str]:
        """
        All account aliases from every account-related plugin. The "alias"
        is part of the :class:`~ape.api.accounts.AccountAPI`. Use the
        account alias to load an account using method
        :meth:`~ape.managers.accounts.AccountManager.load`.

        Returns:
            Iterator[str]
        """

        for container in self.containers.values():
            yield from container.aliases

    def get_accounts_by_type(self, type_: type[AccountAPI]) -> list[AccountAPI]:
        """
        Get a list of accounts by their type.

        Args:
            type_ (type[:class:`~ape.api.accounts.AccountAPI`]): The type of account
              to get.

        Returns:
            list[:class:`~ape.api.accounts.AccountAPI`]
        """

        return [acc for acc in self if isinstance(acc, type_)]

    def __len__(self) -> int:
        """
        The number of accounts managed by all account plugins.

        Returns:
            int
        """
        return sum(len(container) for container in self.containers.values())

    def __iter__(self) -> Iterator[AccountAPI]:
        for container in self.containers.values():
            yield from container.accounts

    @log_instead_of_fail(default="<AccountManager>")
    def __repr__(self) -> str:
        return "<AccountManager>"

    @cached_property
    def test_accounts(self) -> TestAccountManager:
        """
        Accounts generated from the configured test mnemonic. These accounts
        are also the subject of a fixture available in the ``test`` plugin called
        ``accounts``. Configure these accounts, such as the mnemonic and / or
        number-of-accounts using the ``test`` section of the `ape-config.yaml` file.

        Usage example::

            def test_my_contract(accounts):
                # The "accounts" fixture uses the AccountsManager.test_accounts()
                sender = accounts[0]
                receiver = accounts[1]
                ...

        Returns:
            :class:`TestAccountContainer`
        """
        return TestAccountManager()

    def load(self, alias: str) -> AccountAPI:
        """
        Get an account by its alias.

        Raises:
            KeyError: When there is no local account with the given alias.

        Returns:
            :class:`~ape.api.accounts.AccountAPI`
        """
        if alias == "":
            raise ValueError("Cannot use empty string as alias!")

        elif alias in self._alias_to_account_cache:
            return self._alias_to_account_cache[alias]

        for account in self:
            if account.alias and account.alias == alias:
                self._alias_to_account_cache[alias] = account
                return account

        raise KeyError(f"No account with alias '{alias}'.")

    @singledispatchmethod
    def __getitem__(self, account_id) -> AccountAPI:
        raise NotImplementedError(f"Cannot use {type(account_id)} as account ID.")

    @__getitem__.register
    def __getitem_int(self, account_id: int) -> AccountAPI:
        """
        Get an account by index. For example, when you do the CLI command
        ``ape accounts list --all``, you will see a list of enumerated accounts
        by their indices. Use this method as a quicker, ad-hoc way to get an
        account from that index.

        **NOTE**: It is generally preferred to use
        :meth:`~ape.managers.accounts.AccountManager.load` or
        :meth:`~ape.managers.accounts.AccountManager.__getitem_str`.

        Returns:
            :class:`~ape.api.accounts.AccountAPI`
        """
        if account_id < 0:
            account_id = len(self) + account_id
        for idx, account in enumerate(self):
            if account_id == idx:
                return account

        raise IndexError(f"No account at index '{account_id}'.")

    @__getitem__.register
    def __getitem_slice(self, account_id: slice):
        """
        Get list of accounts by slice. For example, when you do the CLI command
        ``ape accounts list --all``, you will see a list of enumerated accounts
        by their indices. Use this method as a quicker, ad-hoc way to get an
        accounts from a slice.

        **NOTE**: It is generally preferred to use
        :meth:`~ape.managers.accounts.AccountManager.load` or
        :meth:`~ape.managers.accounts.AccountManager.__getitem_str`.

        Returns:
            list[:class:`~ape.api.accounts.AccountAPI`]
        """

        start_idx = account_id.start or 0
        if start_idx < 0:
            start_idx += len(self)
        stop_idx = account_id.stop or len(self)
        if stop_idx < 0:
            stop_idx += len(self)
        step_size = account_id.step or 1
        return [self[i] for i in range(start_idx, stop_idx, step_size)]

    @__getitem__.register
    def __getitem_str(self, account_str: str) -> AccountAPI:
        """
        Get an account by address. If we are using a provider that supports unlocking
        accounts, this method will return an impersonated account at that address.

        Raises:
            KeyError: When there is no local account with the given address.

        Returns:
            :class:`~ape.api.accounts.AccountAPI`
        """

        try:
            account_id = self.conversion_manager.convert(account_str, AddressType)
        except ConversionError as err:
            prefix = f"No account with ID '{account_str}'"
            if account_str.endswith(".eth"):
                suffix = "Do you have `ape-ens` installed?"
            else:
                suffix = "Do you have the necessary conversion plugins installed?"

            raise KeyError(f"{prefix}. {suffix}") from err

        for container in self.containers.values():
            if account_id in container.accounts:
                return container[account_id]

        # NOTE: Fallback to `TestAccountContainer`'s method for loading items
        return self.test_accounts[account_id]

    def __contains__(self, address: AddressType) -> bool:
        """
        Determine if the given address matches an account in ``ape``.

        Args:
            address (:class:`~ape.types.address.AddressType`): The address to check.

        Returns:
            bool: ``True`` when the given address is found.
        """
        return (
            any(address in container for container in self.containers.values())
            or address in self.test_accounts
        )

    def use_sender(
        self,
        account_id: Union[AccountAPI, AddressType, str, int],
    ) -> "ContextManager":
        if not isinstance(account_id, AccountAPI):
            if isinstance(account_id, int) or is_hex(account_id):
                account = self[account_id]
            elif isinstance(account_id, str):  # alias
                account = self.load(account_id)
            else:
                raise TypeError(account_id)
        else:
            account = account_id

        return _use_sender(account)

    def init_test_account(
        self, index: int, address: AddressType, private_key: str
    ) -> "TestAccountAPI":
        return self.test_accounts.init_test_account(index, address, private_key)

    def resolve_address(
        self, account_id: Union["BaseAddress", AddressType, str, int, bytes]
    ) -> Optional[AddressType]:
        """
        Resolve the given input to an address.

        Args:
            account_id (:class:~ape.api.address.BaseAddress, str, int, bytes): The input to resolve.
                It handles anything that converts to an AddressType like an ENS or a BaseAddress.
                It also handles account aliases Ape is aware of, or int or bytes address values.

        Returns:
            :class:`~ape.types.AddressType` | None
        """
        if isinstance(account_id, str) and account_id.startswith("0x"):
            # Was given a hex-address string.
            if provider := self.network_manager.active_provider:
                return provider.network.ecosystem.decode_address(account_id)
            else:
                # Assume Ethereum-like.
                return self.network_manager.ether.decode_address(account_id)

        elif not isinstance(account_id, str):
            # Was given either an integer, bytes, or a BaseAddress (account or contract).
            return self.conversion_manager.convert(account_id, AddressType)

        elif isinstance(account_id, str) and account_id in self.aliases:
            # Was given an account alias.
            account = self.load(account_id)
            return account.address

        elif (
            isinstance(account_id, str)
            and account_id.startswith("TEST::")
            and account_id[-1].isdigit()
        ):
            # Test account "alias".
            account_idx = int(account_id[-1])
            return self.test_accounts[account_idx]

        elif isinstance(account_id, str) and not is_hex(account_id):
            # Was maybe given an ENS name.
            try:
                return self.conversion_manager.convert(account_id, AddressType)
            except ConversionError:
                return None

        return None



import pytest

from ape.managers.converters import (
    AddressAPIConverter,
    BytesAddressConverter,
    HexConverter,
    HexIntConverter,
    IntAddressConverter,
    StringDecimalConverter,
    StringIntConverter,
)
from ape.types.address import AddressType
from ape.utils.basemodel import ManagerAccessMixin
from ape_ethereum._converters import WeiConversions

NAME_TO_CONVERTER = {
    "AddressAPI": AddressAPIConverter,
    "BytesAddress": BytesAddressConverter,
    "hex": HexConverter,
    "HexInt": HexIntConverter,
    "IntAddress": IntAddressConverter,
    "StringDecimal": StringDecimalConverter,
    "StringInt": StringIntConverter,
    "wei": WeiConversions,
}


@pytest.fixture
def conversion_manager():
    return ManagerAccessMixin.conversion_manager


@pytest.mark.parametrize("name", NAME_TO_CONVERTER)
def test_get_converter(name, conversion_manager):
    actual = conversion_manager.get_converter(name)
    expected = NAME_TO_CONVERTER[name]
    assert isinstance(actual, expected)


def test_get_converters_by_type(conversion_manager):
    converters = conversion_manager.get_converters_by_type(AddressType)
    for expected in (AddressAPIConverter, IntAddressConverter, BytesAddressConverter):
        assert any(isinstance(c, expected) for c in converters)



"""
Much of the models here are heavily inspired from the rust Alloy crate ``alloy-rpc-types-mev``.
https://github.com/alloy-rs/alloy
"""

from collections.abc import Iterator
from enum import Enum
from typing import Optional, Union

from eth_pydantic_types.hex import HexBytes, HexBytes32, HexInt
from ethpm_types.abi import EventABI
from pydantic import ConfigDict, Field

from ape.exceptions import ProviderNotConnectedError
from ape.types import AddressType
from ape.utils.basemodel import BaseModel, ManagerAccessMixin


class ProtocolVersion(str, Enum):
    """
    The version of the MEV-share API to use.
    """

    BETA1 = "beta-1"
    """
    The beta-1 version of the API.
    """

    V0_1 = "v0.1"
    """
    The 0.1 version of the API.
    """


class Refund(BaseModel):
    """
    Specifies the minimum percent of a given bundle's earnings to redistribute for it to be included
    in a builder's block.
    """

    body_idx: HexInt
    """
    The index of the transaction in the bundle.
    """

    percent: HexInt
    """
    The minimum percent of the bundle's earnings to redistribute.
    """


class PrivacyHint(str, Enum):
    """
    Hints on what data should be shared about the bundle and its transactions.
    """

    CALLDATA = "calldata"
    """
    The calldata of the bundle's transactions should be shared.
    """

    CONTRACT_ADDRESS = "contract_address"
    """
    The address of the bundle's transactions should be shared.
    """

    LOGS = "logs"
    """
    The logs of the bundle's transactions should be shared.
    """

    FUNCTION_SELECTOR = "function_selector"
    """
    The function selector of the bundle's transactions should be shared.
    """

    HASH = "hash"
    """
    The hash of the bundle's transactions should be shared.
    """

    TX_HASH = "tx_hash"
    """
    The hash of the bundle should be shared.
    """


class Privacy(BaseModel):
    """
    Preferences on what data should be shared about the bundle and its transactions
    """

    hints: Optional[list[PrivacyHint]] = None
    """
    Hints on what data should be shared about the bundle and its transactions.
    """

    builders: Optional[list[str]] = None
    """
    Names of the builders that should be allowed to see the bundle/transaction.
    """


class Inclusion(BaseModel):
    """
    Data used by block builders to check if the bundle should be considered for inclusion.
    """

    block: HexInt
    """
    The first block the bundle is valid for.
    """

    max_block: Union[HexInt, None] = Field(None, alias="maxBlock")
    """
    The last block the bundle is valid for.
    """


class BundleHashItem(BaseModel):
    """
    The hash of either a transaction or bundle we are trying to backrun.
    """

    hash: HexBytes32
    """
    Tx hash.
    """


class BundleTxItem(BaseModel):
    """
    A new signed transaction.
    """

    model_config = ConfigDict(populate_by_name=True)

    tx: HexBytes
    """
    Bytes of the signed transaction.
    """

    can_revert: bool = Field(alias="canRevert")
    """
    If true, the transaction can revert without the bundle being considered invalid.
    """


class BundleNestedItem(BaseModel):
    """
    A nested bundle request.
    """

    bundle: "Bundle"
    """
    A bundle request of type Bundle
    """


class RefundConfig(BaseModel):
    """
    Specifies what addresses should receive what percent of the overall refund for this bundle,
    if it is enveloped by another bundle (e.g. a searcher backrun).
    """

    address: AddressType
    """
    The address to refund.
    """

    percent: int
    """
    The minimum percent of the bundle's earnings to redistribute.
    """


class Validity(BaseModel):
    """
    Requirements for the bundle to be included in the block.
    """

    refund: Union[list[Refund], None] = None
    """
    Specifies the minimum percent of a given bundle's earnings to redistribute
    for it to be included in a builder's block.
    """

    refund_config: Optional[list[RefundConfig]] = Field(None, alias="refundConfig")
    """
    Specifies what addresses should receive what percent of the overall refund for this bundle,
    if it is enveloped by another bundle (e.g. a searcher backrun).
    """


class Bundle(BaseModel):
    """
    A bundle of transactions to send to the matchmaker.
    """

    version: ProtocolVersion
    """
    The version of the MEV-share API to use.
    """

    inclusion: Inclusion
    """
    Data used by block builders to check if the bundle should be considered for inclusion.
    """

    body: list[Union[BundleHashItem, BundleTxItem, BundleNestedItem]]
    """
    The transactions to include in the bundle.
    """

    validity: Optional[Validity] = None
    """
    Requirements for the bundle to be included in the block.
    """

    privacy: Optional[Privacy] = None
    """
    Preferences on what data should be shared about the bundle and its transactions
    """

    @classmethod
    def build_for_block(
        cls,
        block: HexInt,
        max_block: Optional[HexInt] = None,
        version: Optional[ProtocolVersion] = None,
        body: Optional[list[Union[BundleHashItem, BundleTxItem, BundleNestedItem]]] = None,
        validity: Optional[Validity] = None,
        privacy: Optional[Privacy] = None,
    ) -> "Bundle":
        return cls(
            version=version or ProtocolVersion.V0_1,
            inclusion=Inclusion(block=block, max_block=max_block),
            body=body or [],
            validity=validity,
            privacy=privacy,
        )

    def add_tx(self, tx: HexBytes, can_revert: bool) -> "Bundle":
        self.body.append(BundleTxItem(tx=tx, can_revert=can_revert))
        return self

    def add_hash(self, hash: HexBytes32) -> "Bundle":
        self.body.append(BundleHashItem(hash=hash))
        return self

    def add_bundle(self, bundle: "Bundle"):
        self.body.append(BundleNestedItem(bundle=bundle))


class SimBundleLogs(BaseModel):
    """
    Logs returned by `mev_simBundle`.
    """

    tx_logs: Optional[list[dict]] = Field(None, alias="txLogs")
    """
    Logs for transactions in bundle.
    """

    bundle_logs: Optional[list["SimBundleLogs"]] = Field(None, alias="bundleLogs")
    """
    Logs for bundles in bundle.
    """


class SimulationReport(BaseModel):
    """
    Response from the matchmaker after sending a simulation request.
    """

    success: bool
    """
    Whether the simulation was successful.
    """

    error: Optional[str] = None
    """
    Error message if the simulation failed.
    """

    state_block: Optional[HexInt] = Field(None, alias="stateBlock")
    """
    The block number of the simulated block.
    """

    mev_gas_price: HexInt = Field(alias="mevGasPrice")
    """
    The profit of the simulated block.
    """

    profit: HexInt
    """
    The profit of the simulated block.
    """

    refundable_value: Optional[HexInt] = Field(None, alias="refundableValue")
    """
    The refundable value of the simulated block.
    """

    gas_used: Optional[HexInt] = Field(None, alias="gasUsed")
    """
    The gas used by the simulated block.
    """

    logs: Optional[list[SimBundleLogs]] = None
    """
    Logs returned by `mev_simBundle`.
    """

    exec_error: Optional[str] = Field(None, alias="execError")
    """
    Error message if the bundle execution failed.
    """

    revert: Optional[HexBytes] = None
    """
    Contains the return data if the transaction reverted
    """

    def decode_logs(self, *events: EventABI):
        try:
            ecosystem = ManagerAccessMixin.provider.network.ecosystem
        except ProviderNotConnectedError:
            # Assume Ethereum (since we are in ape-ethereum after all).
            ecosystem = ManagerAccessMixin.network_manager.ethereum

        return ecosystem.decode_logs(list(self.transaction_logs), *events)

    @property
    def transaction_logs(self, *events: EventABI) -> Iterator[dict]:
        yield from _get_transaction_logs_from_sim_logs(self.logs or [])


def _get_transaction_logs_from_sim_logs(logs: list[SimBundleLogs]) -> Iterator[dict]:
    for bundle_log in logs:
        yield from (bundle_log.tx_logs or [])
        yield from _get_transaction_logs_from_sim_logs(bundle_log.bundle_logs or [])



import os

from tests.integration.cli.utils import run_once

"""
The purpose of this unit test is to test the functionality of `ape init`

It should test the creation of directories, config.yaml, and .gitginore:
contracts
test
scripts
ape-config.yaml
.gitignore
"""


@run_once
def test_init_success(ape_cli, runner, integ_project):
    # Successful creation of project
    # ape init command

    # Changes cwd to a temporary directory
    project_folder_path = integ_project.path / "init_success"
    project_folder_path.mkdir()
    start_path = os.curdir
    os.chdir(str(project_folder_path))

    try:
        result = runner.invoke(ape_cli, ("init",), input="\n".join(["init_success"]))

        assert result.exit_code == 0, result.output
        # checks if the directory exist
        for folder_name in ["contracts", "tests", "scripts"]:
            folder = project_folder_path / folder_name
            assert folder.is_dir()

        # checks if the files exist
        git_ignore_file = project_folder_path / ".gitignore"
        assert git_ignore_file.is_file()
        assert ".env" in git_ignore_file.read_text()

        config = project_folder_path / "ape-config.yaml"
        assert config.is_file()
        assert "init_success" in config.read_text()

    finally:
        os.chdir(start_path)


@run_once
def test_fail_all_files_and_folders_exist(ape_cli, runner, integ_project):
    # failed to create all folders because they exist
    # ape init command

    # add project folder and directories
    start_path = os.curdir
    project_folder_path = integ_project.path / "init_fail"
    project_folder_path.mkdir()
    os.chdir(str(project_folder_path))

    try:
        for folder_name in ["contracts", "tests", "scripts"]:
            # Create target Directory
            folder = project_folder_path / folder_name
            if not folder.exists():
                folder.mkdir(exist_ok=False)

        result = runner.invoke(ape_cli, ("init",), input="\n".join(["init_fail"]))
        # checks if the directory existence
        assert result.exit_code == 0, result.output
        assert "contracts' exists" in result.output
        assert "scripts' exists" in result.output
        assert "tests' exists" in result.output

    finally:
        os.chdir(start_path)
